import json
import pytest

from boxsdk.exception import BoxAPIException
from boxsdk.config import API
from boxsdk.object.watermark import Watermark
from boxsdk.object.collaboration import Collaboration
from boxsdk.exception import BoxValueError


@pytest.fixture(params=('file', 'folder'))
def test_item_and_response(test_file, test_folder, mock_file_response, mock_folder_response, request):
    if request.param == 'file':
        return test_file, mock_file_response
    return test_folder, mock_folder_response


def test_update_info(test_item_and_response, mock_box_session, etag, if_match_header):
    # pylint:disable=redefined-outer-name, protected-access
    test_item, mock_item_response = test_item_and_response
    expected_url = test_item.get_url()
    mock_box_session.put.return_value = mock_item_response
    data = {'foo': 'bar', 'baz': {'foo': 'bar'}, 'num': 4}
    update_response = test_item.update_info(data=data, etag=etag)
    mock_box_session.put.assert_called_once_with(expected_url, data=json.dumps(data), headers=if_match_header, params=None)
    assert isinstance(update_response, test_item.__class__)
    assert update_response.object_id == test_item.object_id


def test_update_info_with_default_request_kwargs(test_item_and_response, mock_box_session, mock_box_session_2):
    # pylint:disable=redefined-outer-name, protected-access
    test_item, mock_item_response = test_item_and_response
    expected_url = test_item.get_url()
    mock_box_session.with_default_network_request_kwargs.return_value = mock_box_session_2
    mock_box_session_2.put.return_value = mock_item_response
    data = {'foo': 'bar', 'baz': {'foo': 'bar'}, 'num': 4}
    extra_network_parameters = {'timeout': 1}
    update_response = test_item.update_info(data=data, extra_network_parameters=extra_network_parameters)
    mock_box_session.with_default_network_request_kwargs.assert_called_once_with({'timeout': 1})
    mock_box_session_2.put.assert_called_once_with(expected_url, data=json.dumps(data), headers=None, params=None)
    assert isinstance(update_response, test_item.__class__)
    assert update_response.object_id == test_item.object_id


def test_clear_unshared_at_for_shared_link(
        test_item_and_response,
        mock_box_session,
        test_url,
):
    test_item, _ = test_item_and_response
    expected_url = test_item.get_url()
    mock_box_session.put.return_value.json.return_value = {
        'type': test_item.object_type,
        'id': test_item.object_id,
        'shared_link': {
            'url': test_url,
            'unshared_at': None,
        },
    }
    expected_data = {'shared_link': {'unshared_at': None, }, }
    shared_link = test_item.get_shared_link(unshared_at=None)
    mock_box_session.put.assert_called_once_with(
        expected_url,
        data=json.dumps(expected_data),
        headers=None,
        params=None,
    )
    assert shared_link is test_url


def test_remove_shared_link(test_item_and_response, mock_box_session, etag, if_match_header):
    # pylint:disable=redefined-outer-name, protected-access
    test_item, _ = test_item_and_response
    expected_url = test_item.get_url()
    mock_box_session.put.return_value.json.return_value = {
        'type': test_item.object_type,
        'id': test_item.object_id,
        'shared_link': None,
    }
    removed = test_item.remove_shared_link(etag=etag)
    mock_box_session.put.assert_called_once_with(
        expected_url,
        data=json.dumps({'shared_link': None}),
        headers=if_match_header,
        params=None,
    )
    assert removed is True


@pytest.mark.parametrize('fields', (None, ['name', 'created_at']))
def test_get(test_item_and_response, mock_box_session, fields, mock_object_id, etag, if_none_match_header):
    # pylint:disable=redefined-outer-name, protected-access
    test_item, mock_item_response = test_item_and_response
    expected_url = test_item.get_url()
    mock_box_session.get.return_value = mock_item_response
    expected_params = {'fields': ','.join(fields)} if fields else None
    info = test_item.get(fields=fields, etag=etag)
    mock_box_session.get.assert_called_once_with(expected_url, params=expected_params, headers=if_none_match_header)
    assert isinstance(info, test_item.__class__)
    assert info.id == mock_object_id


def test_get_watermark(test_item_and_response, mock_box_session):
    test_item, _ = test_item_and_response
    created_at = '2016-10-31T15:33:33-07:00'
    modified_at = '2016-10-31T15:33:33-07:00'
    expected_url = f'{API.BASE_API_URL}/{test_item.object_type}s/{test_item.object_id}/watermark'
    mock_box_session.get.return_value.json.return_value = {
        'watermark': {
            'created_at': created_at,
            'modified_at': modified_at,
        },
    }
    watermark = test_item.get_watermark()
    mock_box_session.get.assert_called_once_with(expected_url)
    assert isinstance(watermark, Watermark)
    assert watermark['created_at'] == created_at
    assert watermark['modified_at'] == modified_at


def test_apply_watermark(test_item_and_response, mock_box_session):
    test_item, _ = test_item_and_response
    created_at = '2016-10-31T15:33:33-07:00'
    modified_at = '2016-10-31T15:33:33-07:00'
    expected_url = f'{API.BASE_API_URL}/{test_item.object_type}s/{test_item.object_id}/watermark'
    mock_box_session.put.return_value.json.return_value = {
        'watermark': {
            'created_at': created_at,
            'modified_at': modified_at,
        },
    }
    watermark = test_item.apply_watermark()
    mock_box_session.put.assert_called_once_with(expected_url, data='{"watermark": {"imprint": "default"}}')
    assert isinstance(watermark, Watermark)
    assert watermark['created_at'] == created_at
    assert watermark['modified_at'] == modified_at


def test_delete_watermark(test_item_and_response, mock_box_session):
    test_item, _ = test_item_and_response
    expected_url = f'{API.BASE_API_URL}/{test_item.object_type}s/{test_item.object_id}/watermark'
    mock_box_session.delete.return_value.ok = True
    is_watermark_deleted = test_item.delete_watermark()
    mock_box_session.delete.assert_called_once_with(expected_url, expect_json_response=False)
    assert is_watermark_deleted is True


def test_collaborate_with_group(test_item_and_response, test_group, mock_box_session):
    # pylint:disable=redefined-outer-name, protected-access
    test_item, _ = test_item_and_response
    expected_url = f'{API.BASE_API_URL}/collaborations'
    expected_data = {
        'item': {
            'type': test_item.object_type,
            'id': test_item.object_id,
        },
        'accessible_by': {
            'type': test_group.object_type,
            'id': test_group.object_id,
        },
        'role': 'editor',
    }
    mock_collaboration = {
        'type': 'collaboration',
        'id': '1234',
        'created_by': {
            'type': 'user',
            'id': '1111',
        }
    }
    mock_box_session.post.return_value.json.return_value = mock_collaboration
    collaboration = test_item.collaborate(test_group, 'editor')
    mock_box_session.post.assert_called_once_with(expected_url, data=json.dumps(expected_data), params={})
    assert collaboration.id == mock_collaboration['id']
    assert collaboration['type'] == mock_collaboration['type']
    assert collaboration['created_by']['id'] == mock_collaboration['created_by']['id']


@pytest.mark.parametrize('can_view_path,fields,notify,data,params', [
    (None, None, None, {}, {}),
    (True, None, None, {'can_view_path': True}, {}),
    (False, None, None, {'can_view_path': False}, {}),
    (None, ['type', 'id', 'created_by'], None, {}, {'fields': 'type,id,created_by'}),
    (None, None, True, {}, {'notify': True}),
    (None, None, False, {}, {'notify': False}),
    (True, ['type', 'id', 'created_by'], False, {'can_view_path': True}, {'fields': 'type,id,created_by', 'notify': False})
])
def test_collaborate_with_user(test_item_and_response, mock_user, mock_box_session, can_view_path, fields, notify, data, params):
    # pylint:disable=redefined-outer-name, protected-access
    test_item, _ = test_item_and_response
    expected_url = f'{API.BASE_API_URL}/collaborations'
    expected_data = {
        'item': {
            'type': test_item.object_type,
            'id': test_item.object_id,
        },
        'accessible_by': {
            'type': mock_user.object_type,
            'id': mock_user.object_id,
        },
        'role': 'editor',
    }
    expected_data.update(data)
    mock_collaboration = {
        'type': 'collaboration',
        'id': '1234',
        'created_by': {
            'type': 'user',
            'id': '1111',
        }
    }
    expected_params = params
    mock_box_session.post.return_value.json.return_value = mock_collaboration
    collaboration = test_item.collaborate(mock_user, 'editor', can_view_path=can_view_path, fields=fields, notify=notify)
    mock_box_session.post.assert_called_once_with(expected_url, data=json.dumps(expected_data), params=expected_params)
    assert collaboration.id == mock_collaboration['id']
    assert collaboration['type'] == mock_collaboration['type']
    assert collaboration['created_by']['id'] == mock_collaboration['created_by']['id']


@pytest.mark.parametrize('can_view_path,fields,notify,data,params', [
    (None, None, None, {}, {}),
    (True, None, None, {'can_view_path': True}, {}),
    (False, None, None, {'can_view_path': False}, {}),
    (None, ['type', 'id', 'created_by'], None, {}, {'fields': 'type,id,created_by'}),
    (None, None, True, {}, {'notify': True}),
    (None, None, False, {}, {'notify': False}),
    (True, ['type', 'id', 'created_by'], False, {'can_view_path': True}, {'fields': 'type,id,created_by', 'notify': False})
])
def test_collaborate_with_login(test_item_and_response, mock_box_session, can_view_path, fields, notify, data, params):
    # pylint:disable=redefined-outer-name, protected-access
    test_item, _ = test_item_and_response
    expected_url = f'{API.BASE_API_URL}/collaborations'
    expected_data = {
        'item': {
            'type': test_item.object_type,
            'id': test_item.object_id,
        },
        'accessible_by': {
            'type': 'user',
            'login': 'test@example.com',
        },
        'role': 'editor',
    }
    expected_data.update(data)
    mock_collaboration = {
        'type': 'collaboration',
        'id': '1234',
        'created_by': {
            'type': 'user',
            'id': '1111',
        }
    }
    expected_params = params
    mock_box_session.post.return_value.json.return_value = mock_collaboration
    collaboration = test_item.collaborate_with_login('test@example.com', 'editor', can_view_path=can_view_path, fields=fields, notify=notify)
    mock_box_session.post.assert_called_once_with(expected_url, data=json.dumps(expected_data), params=expected_params)
    assert collaboration.id == mock_collaboration['id']
    assert collaboration['type'] == mock_collaboration['type']
    assert collaboration['created_by']['id'] == mock_collaboration['created_by']['id']


def test_collaborations(test_item_and_response, mock_box_session):
    # pylint:disable=redefined-outer-name, protected-access
    test_item, _ = test_item_and_response
    expected_url = f'{API.BASE_API_URL}/{test_item.object_type}s/{test_item.object_id}/collaborations'
    mock_collaboration = {
        'type': 'collaboration',
        'id': '12345',
        'created_by': {
            'type': 'user',
            'id': '33333',
        },
    }
    mock_box_session.get.return_value.json.return_value = {
        'limit': 500,
        'entries': [mock_collaboration]
    }
    collaborations = test_item.get_collaborations(limit=500)
    collaboration = collaborations.next()
    mock_box_session.get.assert_called_once_with(expected_url, params={'limit': 500})
    assert isinstance(collaboration, Collaboration)
    assert collaboration.id == mock_collaboration['id']
    assert collaboration.type == mock_collaboration['type']
    assert collaboration['created_by']['type'] == 'user'
    assert collaboration['created_by']['id'] == '33333'


def test_get_all_metadata(test_item_and_response, mock_box_session):
    test_item, _ = test_item_and_response
    expected_url = f'{API.BASE_API_URL}/{test_item.object_type}s/{test_item.object_id}/metadata'
    mock_metadata = {
        'currentDocumentStage': 'prioritization',
        'needsApprovalFrom': 'planning team',
        '$type': 'documentFlow-452b4c9d-c3ad-4ac7-b1ad-9d5192f2fc5f',
        '$parent': 'folder_998951261',
        '$id': 'e57f90ff-0044-48c2-807d-06b908765baf',
        '$version': 1,
        '$typeVersion': 2,
        'maximumDaysAllowedInCurrentStage': 5,
        '$template': 'documentFlow',
        '$scope': 'enterprise_12345',
    }
    mock_box_session.get.return_value.json.return_value = {
        'limit': 100,
        'entries': [mock_metadata]
    }

    all_metadata = test_item.get_all_metadata()
    metadata = all_metadata.next()

    mock_box_session.get.assert_called_once_with(expected_url, params={})
    assert isinstance(metadata, dict)
    for key in metadata:
        assert mock_metadata[key] == mock_metadata[key]


def test_add_classification(test_item_and_response, mock_box_session):
    # pylint:disable=redefined-outer-name
    test_item, _ = test_item_and_response
    expected_url = f'{API.BASE_API_URL}/{test_item.object_type}s/{test_item.object_id}' \
                   f'/metadata/enterprise/securityClassification-6VMVochwUWo'
    metadata_response = {
        'Box__Security__Classification__Key': 'Public',
    }
    metadata_response = mock_box_session.post.return_value.json.return_value = metadata_response
    data = {
        'Box__Security__Classification__Key': 'Public'
    }
    headers = {
        b'Content-Type': b'application/json'
    }
    metadata = test_item.add_classification('Public')
    mock_box_session.post.assert_called_once_with(expected_url, headers=headers, data=json.dumps(data))
    assert metadata is metadata_response['Box__Security__Classification__Key']


def test_update_classification(test_item_and_response, mock_box_session):
    # pylint:disable=redefined-outer-name
    test_item, _ = test_item_and_response
    expected_url = f'{API.BASE_API_URL}/{test_item.object_type}s/{test_item.object_id}' \
                   f'/metadata/enterprise/securityClassification-6VMVochwUWo'
    metadata_response = {
        'Box__Security__Classification__Key': 'Internal',
    }
    metadata_response = mock_box_session.put.return_value.json.return_value = metadata_response
    data = [{
        'op': 'add',
        'path': '/Box__Security__Classification__Key',
        'value': 'Internal',
    }]
    headers = {
        b'Content-Type': b'application/json-patch+json'
    }
    metadata = test_item.update_classification('Internal')
    mock_box_session.put.assert_called_once_with(expected_url, headers=headers, data=json.dumps(data))
    assert metadata is metadata_response['Box__Security__Classification__Key']


def test_set_classification_succeeds(test_item_and_response, mock_box_session):
    # pylint:disable=redefined-outer-name
    test_item, _ = test_item_and_response
    metadata_response = {
        'Box__Security__Classification__Key': 'Public',
    }
    expected_url = f'{API.BASE_API_URL}/{test_item.object_type}s/{test_item.object_id}' \
                   f'/metadata/enterprise/securityClassification-6VMVochwUWo'
    post_data = {
        'Box__Security__Classification__Key': 'Public',
    }
    put_data = [{
        'op': 'add',
        'path': '/Box__Security__Classification__Key',
        'value': 'Public',
    }]
    post_headers = {
        b'Content-Type': b'application/json'
    }
    put_headers = {
        b'Content-Type': b'application/json-patch+json'
    }
    mock_box_session.post.side_effect = [BoxAPIException(status=409)]
    mock_box_session.put.return_value.json.return_value = metadata_response
    metadata = test_item.set_classification('Public')
    mock_box_session.post.assert_called_once_with(expected_url, data=json.dumps(post_data), headers=post_headers)
    mock_box_session.put.assert_called_once_with(expected_url, data=json.dumps(put_data), headers=put_headers)
    assert metadata is metadata_response['Box__Security__Classification__Key']


def test_set_classification_fails(test_item_and_response, mock_box_session):
    # pylint:disable=redefined-outer-name
    test_item, _ = test_item_and_response
    expected_url = f'{API.BASE_API_URL}/{test_item.object_type}s/{test_item.object_id}' \
                   f'/metadata/enterprise/securityClassification-6VMVochwUWo'
    post_data = {
        'Box__Security__Classification__Key': 'Public',
    }
    post_headers = {
        b'Content-Type': b'application/json'
    }
    mock_box_session.post.side_effect = [BoxAPIException(status=500)]
    with pytest.raises(BoxAPIException):
        test_item.set_classification('Public')
    mock_box_session.post.assert_called_once_with(expected_url, data=json.dumps(post_data), headers=post_headers)


def test_get_classification_succeeds(test_item_and_response, mock_box_session):
    # pylint:disable=redefined-outer-name
    test_item, _ = test_item_and_response
    expected_url = f'{API.BASE_API_URL}/{test_item.object_type}s/{test_item.object_id}' \
                   f'/metadata/enterprise/securityClassification-6VMVochwUWo'
    metadata_response = {
        'Box__Security__Classification__Key': 'Public'
    }
    mock_box_session.get.return_value.json.return_value = metadata_response
    metadata = test_item.get_classification()
    assert metadata is metadata_response['Box__Security__Classification__Key']
    mock_box_session.get.assert_called_once_with(expected_url)


def test_get_classification_not_found(test_item_and_response, mock_box_session):
    # pylint:disable=redefined-outer-name
    test_item, _ = test_item_and_response
    expected_url = f'{API.BASE_API_URL}/{test_item.object_type}s/{test_item.object_id}' \
                   f'/metadata/enterprise/securityClassification-6VMVochwUWo'
    mock_box_session.get.side_effect = [BoxAPIException(status=404, code="instance_not_found")]
    metadata = test_item.get_classification()
    assert metadata is None
    mock_box_session.get.assert_called_once_with(expected_url)


def test_get_classification_raises_exception(test_item_and_response, mock_box_session):
    # pylint:disable=redefined-outer-name
    test_item, _ = test_item_and_response
    expected_url = f'{API.BASE_API_URL}/{test_item.object_type}s/{test_item.object_id}' \
                   f'/metadata/enterprise/securityClassification-6VMVochwUWo'
    mock_box_session.get.side_effect = [BoxAPIException(status=500)]
    with pytest.raises(BoxAPIException):
        test_item.get_classification()
    mock_box_session.get.assert_called_once_with(expected_url)


def test_remove_classification(test_item_and_response, mock_box_session, make_mock_box_request):
    # pylint:disable=redefined-outer-name
    test_item, _ = test_item_and_response
    expected_url = f'{API.BASE_API_URL}/{test_item.object_type}s/{test_item.object_id}' \
                   f'/metadata/enterprise/securityClassification-6VMVochwUWo'
    mock_box_session.delete.return_value, _ = make_mock_box_request(response_ok='success')
    is_removed = test_item.remove_classification()
    mock_box_session.delete.assert_called_once_with(expected_url)
    assert is_removed == 'success'


def test_sanitize_item_id(test_item_and_response):
    # pylint:disable=redefined-outer-name, protected-access
    test_item, _ = test_item_and_response
    assert test_item.validate_item_id(test_item._object_id) is None
    test_item._object_id = "foo"
    with pytest.raises(BoxValueError):
        test_item.validate_item_id(test_item._object_id)



from unittest.mock import patch


def test_too_many_requests_causes_retry(box_client, mock_box, monkeypatch):
    monkeypatch.setattr(mock_box, 'RATE_LIMIT_THRESHOLD', 1)
    with patch('random.uniform', return_value=1):
        box_client.folder('0').get()
        box_client.folder('0').get()
    assert len(mock_box.requests) == 6  # 3 auth requests, 2 real requests, and a retry



from abc import ABCMeta
from collections.abc import Mapping

from .base_api_json_object import BaseAPIJSONObject, BaseAPIJSONObjectMeta


class APIJSONObjectMeta(BaseAPIJSONObjectMeta, ABCMeta):
    """
    Avoid conflicting metaclass definitions for APIJSONObject.
    http://code.activestate.com/recipes/204197-solving-the-metaclass-conflict/
    """


class APIJSONObject(BaseAPIJSONObject, Mapping, metaclass=APIJSONObjectMeta):
    """Class representing objects that are not part of the REST API."""

    def __len__(self) -> int:
        return len(self._response_object)



from typing import Any

from .base_object import BaseObject


class TermsOfServiceUserStatus(BaseObject):
    """Represents a Box terms of service user status."""

    _item_type = 'terms_of_service_user_status'

    def get_url(self, *args: Any) -> str:
        return self._session.get_url('terms_of_service_user_statuses', self._object_id, *args)

    def accept(self) -> 'TermsOfServiceUserStatus':
        """
        Accept a term of service.
        """
        body = {
            'is_accepted': True
        }
        return self.update_info(data=body)

    def reject(self) -> 'TermsOfServiceUserStatus':
        """
        Reject a term of service.
        """
        body = {
            'is_accepted': False
        }
        return self.update_info(data=body)



from typing import Callable, Any, Tuple

from .oauth2 import OAuth2


class RemoteOAuth2Mixin(OAuth2):
    """
    Box SDK OAuth2 mixin.
    Allows for storing auth tokens remotely.

    """
    def __init__(self, retrieve_access_token: Callable[[str], str] = None, *args: Any, **kwargs: Any):
        """
        :param retrieve_access_token:
            Callback to exchange an existing access token for a new one.
        """
        # pylint:disable=keyword-arg-before-vararg
        self._retrieve_access_token = retrieve_access_token
        super().__init__(*args, **kwargs)

    def _refresh(self, access_token: str) -> Tuple[str, None]:
        """
        Base class override. Ask the remote host for a new token.
        """
        self._access_token = self._retrieve_access_token(access_token)
        return self._access_token, None


class RemoteOAuth2(RemoteOAuth2Mixin):
    """
    Box SDK OAuth2 subclass.
    Allows for storing auth tokens remotely. The retrieve_access_token callback should
    return an access token, presumably acquired from a remote server on which your auth credentials are available.
    """



import json

from boxsdk.config import API
from boxsdk.object.retention_policy import RetentionPolicy
from boxsdk.object.retention_policy_assignment import RetentionPolicyAssignment


def test_get(test_retention_policy, mock_box_session):
    expected_url = f'{API.BASE_API_URL}/retention_policies/{test_retention_policy.object_id}'
    mock_box_session.get.return_value.json.return_value = {
        'type': test_retention_policy.object_type,
        'id': test_retention_policy.object_id,
        'policy_name': 'Policy Name',
        'policy_type': 'finite',
        'retention_length': '10',
        'disposition_action': 'permanently_delete',
        'description': 'Description',
    }
    retention_policy = test_retention_policy.get()
    mock_box_session.get.assert_called_once_with(expected_url, headers=None, params=None)
    assert isinstance(retention_policy, RetentionPolicy)
    assert retention_policy['type'] == test_retention_policy.object_type
    assert retention_policy['id'] == test_retention_policy.object_id
    assert retention_policy['policy_name'] == 'Policy Name'
    assert retention_policy['description'] == 'Description'


def test_update(test_retention_policy, mock_box_session):
    new_policy_name = 'New Name'
    new_retention_type = 'modifiable'
    new_retention_length = 60
    new_description = 'New Description'
    expected_url = f'{API.BASE_API_URL}/retention_policies/{test_retention_policy.object_id}'
    mock_box_session.put.return_value.json.return_value = {
        'type': test_retention_policy.object_type,
        'id': test_retention_policy.object_id,
        'policy_name': new_policy_name,
        'policy_type': 'finite',
        'retention_length': new_retention_length,
        'retention_type': new_retention_type,
        'description': new_description,
    }
    data = {
        'policy_name': new_policy_name,
        'retention_type': new_retention_type,
        'retention_length': new_retention_length,
        'description': new_description,
    }
    retention_policy = test_retention_policy.update_info(data=data)
    mock_box_session.put.assert_called_once_with(expected_url, data=json.dumps(data), headers=None, params=None)
    assert isinstance(retention_policy, RetentionPolicy)
    assert retention_policy['type'] == test_retention_policy.object_type
    assert retention_policy['id'] == test_retention_policy.object_id
    assert retention_policy['policy_name'] == new_policy_name
    assert retention_policy['retention_type'] == new_retention_type
    assert retention_policy['retention_length'] == new_retention_length
    assert retention_policy['description'] == new_description


def test_assign(test_retention_policy, test_folder, mock_box_session):
    policy_id = '42'
    expected_url = mock_box_session.get_url('retention_policy_assignments')
    expected_data = {
        'policy_id': policy_id,
        'assign_to': {
            'type': test_folder.object_type,
            'id': test_folder.object_id,
        },
        "start_date_field": "upload_date"
    }
    mock_assignment = {
        'type': 'retention_policy_assignment',
        'id': '1234',
        'retention_policy': {
            'type': 'retention_policy',
            'id': policy_id,
        },
        "start_date_field": "upload_date"
    }
    mock_box_session.post.return_value.json.return_value = mock_assignment
    assignment = test_retention_policy.assign(test_folder, start_date_field='upload_date')
    mock_box_session.post.assert_called_once_with(expected_url, data=json.dumps(expected_data), params={})
    assert assignment.id == mock_assignment['id']
    assert assignment.retention_policy['id'] == mock_assignment['retention_policy']['id']
    assert assignment.start_date_field == mock_assignment['start_date_field']
    assert isinstance(assignment, RetentionPolicyAssignment)


def test_get_assignments(test_retention_policy, mock_box_session):
    expected_url = test_retention_policy.get_url('assignments')
    mock_assignment = {
        'type': 'retention_policy_assignment',
        'id': '12345',
    }
    mock_box_session.get.return_value.json.return_value = {
        'limit': 100,
        'entries': [mock_assignment],
        'next_marker': 'testMarker',
    }
    assignments = test_retention_policy.assignments()
    assignment = assignments.next()
    mock_box_session.get.assert_called_once_with(expected_url, params={})
    assert isinstance(assignment, RetentionPolicyAssignment)
    assert assignment.id == mock_assignment['id']
    assert assignment.type == mock_assignment['type']



from .base_object import BaseObject


class GroupMembership(BaseObject):
    """Represents a Box group_membership, which relates a user & group under a specific role."""

    _item_type = 'group_membership'



import json
from typing import Any, TYPE_CHECKING
from .base_object import BaseObject

if TYPE_CHECKING:
    from boxsdk.object.user import User
    from boxsdk.object.storage_policy_assignment import StoragePolicyAssignment


class StoragePolicy(BaseObject):
    """Represents the storage policy"""

    _item_type = 'storage_policy'

    def get_url(self, *args: Any) -> str:
        """
        Get url for storage policies.
        """
        return self._session.get_url('storage_policies', self._object_id, *args)

    def assign(self, user: 'User') -> 'StoragePolicyAssignment':
        """
        Checks to see if a user is already assigned a storage policy or if the storage policy assigned
        to user belongs to the enterprise. If neither, then update the user storage policy to the new one.

        :param user:
            The class:`User` to assign the storage policy to
        :returns:
            Information about the :class:`StoragePolicyAssignment` object.
        """
        assignment = user.get_storage_policy_assignment()
        if assignment.id == self.object_id:
            return assignment

        if assignment.assigned_to['type'] == 'enterprise':
            return self.create_assignment(user)

        update_object = {
            'storage_policy': {
                'type': self.object_type,
                'id': self.object_id,
            },
        }
        return assignment.update_info(data=update_object)

    def create_assignment(self, user: 'User') -> 'StoragePolicyAssignment':
        """
        Assign a storage policy to a :class:`User`.

        :param user:
            The :class:'User` to assign the storage policy to.
        :returns:
            Information about the :class:`StoragePolicyAssignment` object
        """
        url = self._session.get_url('storage_policy_assignments')
        body = {
            'storage_policy': {
                'type': 'storage_policy',
                'id': self.object_id,
            },
            'assigned_to': {
                'type': user.object_type,
                'id': user.object_id,
            }
        }
        response = self._session.post(url, data=json.dumps(body)).json()
        return self.translator.translate(
            session=self._session,
            response_object=response,
        )



from itertools import chain, islice, repeat, count
import json
from operator import sub
from unittest.mock import Mock

import pytest

from boxsdk.network.default_network import DefaultNetworkResponse
from boxsdk.object.collaboration import Collaboration
from boxsdk.object.group_membership import GroupMembership
from boxsdk.object.user import User
from boxsdk.config import API
from boxsdk.session.box_response import BoxResponse


@pytest.fixture(scope='module')
def delete_group_response():
    # pylint:disable=redefined-outer-name
    mock_network_response = Mock(DefaultNetworkResponse)
    mock_network_response.ok = True
    return mock_network_response


def test_delete_group_return_the_correct_response(
        mock_group,
        mock_box_session,
        delete_group_response,
):
    # pylint:disable=redefined-outer-name
    mock_box_session.delete.return_value = delete_group_response
    response = mock_group.delete()

    # pylint:disable=protected-access
    expected_url = mock_group.get_url()
    # pylint:enable=protected-access
    mock_box_session.delete.assert_called_once_with(expected_url, params={}, expect_json_response=False, headers=None)

    assert response is True


@pytest.mark.parametrize('role', ['member', 'admin'])
def test_add_member(test_group, mock_box_session, mock_add_member_response, mock_user, role):
    expected_url = f'{API.BASE_API_URL}/group_memberships'
    mock_box_session.post.return_value = mock_add_member_response
    new_group_membership = test_group.add_member(mock_user, role, configurable_permissions={'can_run_reports': True})
    data = json.dumps({
        'user': {'id': mock_user.object_id},
        'group': {'id': test_group.object_id},
        'role': role,
        'configurable_permissions': {'can_run_reports': True}
    })
    mock_box_session.post.assert_called_once_with(expected_url, data=data)
    assert isinstance(new_group_membership, GroupMembership)


def test_add_member_default_permission(test_group, mock_box_session, mock_add_member_response, mock_user):
    expected_url = f'{API.BASE_API_URL}/group_memberships'
    mock_box_session.post.return_value = mock_add_member_response
    new_group_membership = test_group.add_member(mock_user, 'member')
    data = json.dumps({
        'user': {'id': mock_user.object_id},
        'group': {'id': test_group.object_id},
        'role': 'member',
    })
    mock_box_session.post.assert_called_once_with(expected_url, data=data)
    assert isinstance(new_group_membership, GroupMembership)


def test_add_member_none_permission(test_group, mock_box_session, mock_add_member_response, mock_user):
    expected_url = f'{API.BASE_API_URL}/group_memberships'
    mock_box_session.post.return_value = mock_add_member_response
    new_group_membership = test_group.add_member(mock_user, 'member', configurable_permissions=None)
    data = json.dumps({
        'user': {'id': mock_user.object_id},
        'group': {'id': test_group.object_id},
        'role': 'member',
        'configurable_permissions': None
    })
    mock_box_session.post.assert_called_once_with(expected_url, data=data)
    assert isinstance(new_group_membership, GroupMembership)


@pytest.fixture()
def mock_membership_dict_stream():
    def gen_data(some_id):
        return {
            'type': 'group_membership',
            'id': f"membership_id_{some_id}",
            'role': 'member',
            'user': {'type': 'user', 'id': f"user_id_{some_id}"},
            'group': {'type': 'group', 'id': f"group_id_{some_id}"},
        }

    return map(gen_data, count())


@pytest.fixture()
def mock_membership_responses(mock_membership_dict_stream):
    """
    Returns a generator method that takes params: total, page_size.
    The generator generates a sequence of 'group membership' mock_box_responses each containing page_size
    items, until 'total' entries have been returned
    """
    # pylint:disable=redefined-outer-name
    def number_entries_per_response(total, page_size, hidden_in_batch):
        if not hidden_in_batch:
            hidden_in_batch = repeat(0)
        quotient, remainder = divmod(total, page_size)
        max_items_in_batch = chain(repeat(page_size, quotient), (remainder,))
        return map(sub, max_items_in_batch, hidden_in_batch)

    def take(iterable, number):
        return list(islice(iterable, number))

    def membership_responses(total, page_size, hidden_in_batch=None):
        offset = 0
        for number_entries in number_entries_per_response(total, page_size, hidden_in_batch):
            entries = take(mock_membership_dict_stream, number_entries)

            mock_box_response = Mock(BoxResponse)
            mock_network_response = Mock(DefaultNetworkResponse)
            mock_box_response.network_response = mock_network_response
            mock_box_response.json.return_value = {
                'entries': entries,
                'total_count': total,
                'offset': offset,
                'limit': page_size,
            }
            offset += page_size
            mock_box_response.status_code = 200
            mock_box_response.ok = True
            yield mock_box_response

    return membership_responses


@pytest.mark.parametrize('total, page_size', [
    (0, 6),
    (5, 6),
    (6, 6),
    (5, 4),
    (9, 4),
])
def test_get_memberships(test_group, mock_box_session, mock_membership_responses, total, page_size):
    # pylint:disable=redefined-outer-name
    # Each call the 'get' (the GET next page call) will return the next response
    mock_box_session.get.side_effect = mock_membership_responses(total, page_size)

    # Get all the members
    all_members = test_group.get_memberships()

    # Assert we got the expected number of membership instances
    count = 0
    for membership in all_members:
        count += 1
        assert isinstance(membership, GroupMembership)
    assert count == total


@pytest.mark.parametrize('hidden_in_batch', [
    (1, 0, 0),
    (0, 0, 1),
    (10, 10, 9),
    (10, 10, 10),
])
def test_get_memberships_with_hidden_results(test_group, mock_box_session, mock_membership_responses, hidden_in_batch):
    """
    This test verifies that the SDK properly deals with missing (aka hidden) data potentially present in a paged
    API. The API might indicate that the total_size is X, but in actuality the pages API could return less than X
    because the auth'd user might not have access to all X of the resources.
    """
    # pylint:disable=redefined-outer-name
    total = 30
    page_size = 10

    total_hidden = sum(hidden_in_batch)

    # Each call the 'get' (the GET next page call) will return the next response
    mock_box_session.get.side_effect = mock_membership_responses(total, page_size, hidden_in_batch=hidden_in_batch)

    # Get all the members
    all_members = test_group.get_memberships(limit=page_size, offset=0)

    # Assert we got the expected number of membership instances
    count = 0
    for membership in all_members:
        count += 1
        assert isinstance(membership, GroupMembership)
    assert count == total - total_hidden


def test_get_group_collaborations(test_group, mock_box_session):
    expected_url = f'{API.BASE_API_URL}/groups/{test_group.object_id}/collaborations'
    mock_collaboration = {
        'type': 'collaboration',
        'id': '12345',
        'created_by': {
            'type': 'user',
            'id': '33333'
        }
    }
    mock_box_session.get.return_value.json.return_value = {
        'limit': 100,
        'entries': [mock_collaboration],
        'offset': 0,
        'total_count': 1
    }
    collaborations = test_group.get_collaborations(fields=['type', 'id', 'created_by'])
    collaboration = collaborations.next()
    mock_box_session.get.assert_called_once_with(expected_url, params={'offset': None, 'fields': 'type,id,created_by'})
    assert isinstance(collaboration, Collaboration)
    assert collaboration.id == mock_collaboration['id']
    assert collaboration.created_by['id'] == mock_collaboration['created_by']['id']


def test_base_api_json_object_returns_correctly(test_group_membership, mock_box_session):
    expected_data = {
        'type': 'group_membership',
        'id': '12345',
        'test': [
            {
                'user': {
                    'type': 'user'
                }
            },
            {
                'group': {
                    'type': 'group'
                }
            },
        ],
        'user': {
            'type': 'user',
            'id': '5678',
            'name': 'Test User',
            'login': 'test@example.com',
        },
        'group': {
            'type': 'group',
            'id': '54321',
            'name': 'Test'
        },
        'role': 'admin',
        'configurable_permissions': {
            'can_run_reports': False,
            'can_instant_login': True,
            'can_create_accounts': False,
            'can_edit_accounts': True,
        },
        'created_at': '2013-05-16T15:27:57-07:00',
        'modified_at': '2013-05-16T15:27:57-07:00',
    }
    mock_box_session.get.return_value.json.return_value = expected_data
    membership = test_group_membership.get()
    membership_response = membership.response_object
    assert isinstance(membership.user, User)
    assert isinstance(membership_response, dict)
    assert membership_response is not expected_data
    assert membership_response == expected_data



# pylint: disable=too-many-lines
import json
from datetime import datetime
from typing import TYPE_CHECKING, Optional, Iterable, Union, Any, IO

from ..auth.oauth2 import TokenResponse
from ..session.session import Session, AuthorizedSession
from ..object.cloneable import Cloneable
from ..util.api_call_decorator import api_call
from ..object.search import Search
from ..object.events import Events
from ..object.collaboration_allowlist import CollaborationAllowlist
from ..object.trash import Trash
from ..pagination.limit_offset_based_object_collection import LimitOffsetBasedObjectCollection
from ..pagination.marker_based_object_collection import MarkerBasedObjectCollection
from ..util.datetime_formatter import normalize_date_to_rfc3339_format
from ..util.shared_link import get_shared_link_header
from ..util.deprecation_decorator import deprecated

if TYPE_CHECKING:
    from boxsdk import OAuth2
    from boxsdk.util.translator import Translator
    from boxsdk.object.folder import Folder
    from boxsdk.object.file import File
    from boxsdk.object.file_request import FileRequest
    from boxsdk.object.file_version import FileVersion
    from boxsdk.object.upload_session import UploadSession
    from boxsdk.object.comment import Comment
    from boxsdk.object.legal_hold import LegalHold
    from boxsdk.object.legal_hold_policy_assignment import LegalHoldPolicyAssignment
    from boxsdk.object.legal_hold_policy import LegalHoldPolicy
    from boxsdk.object.collaboration_allowlist_exempt_target import CollaborationAllowlistExemptTarget
    from boxsdk.object.collaboration_allowlist_entry import CollaborationAllowlistEntry
    from boxsdk.object.collaboration import Collaboration
    from boxsdk.object.group import Group
    from boxsdk.object.email_alias import EmailAlias
    from boxsdk.object.invite import Invite
    from boxsdk.object.user import User
    from boxsdk.object.sign_request import SignRequest
    from boxsdk.object.folder_lock import FolderLock
    from boxsdk.object.metadata_template import MetadataTemplate, MetadataField
    from boxsdk.object.metadata_cascade_policy import MetadataCascadePolicy
    from boxsdk.object.device_pinner import DevicePinner
    from boxsdk.auth.oauth2 import TokenScope
    from boxsdk.session.box_response import BoxResponse
    from boxsdk.object.item import Item
    from boxsdk.object.web_link import WebLink
    from boxsdk.object.retention_policy_assignment import RetentionPolicyAssignment
    from boxsdk.object.file_version_retention import FileVersionRetention
    from boxsdk.object.retention_policy import RetentionPolicy
    from boxsdk.object.task_assignment import TaskAssignment
    from boxsdk.object.task import Task
    from boxsdk.object.terms_of_service_user_status import TermsOfServiceUserStatus
    from boxsdk.object.terms_of_service import TermsOfService, TermsOfServiceType, TermsOfServiceStatus
    from boxsdk.object.storage_policy_assignment import StoragePolicyAssignment
    from boxsdk.object.storage_policy import StoragePolicy
    from boxsdk.object.webhook import Webhook
    from boxsdk.object.group_membership import GroupMembership
    from boxsdk.object.enterprise import Enterprise
    from boxsdk.object.collection import Collection
    from boxsdk.object.sign_template import SignTemplate
    from boxsdk.pagination.box_object_collection import BoxObjectCollection


class Client(Cloneable):
    unauthorized_session_class = Session
    authorized_session_class = AuthorizedSession

    def __init__(self, oauth: 'OAuth2', session: Session = None):
        """
        :param oauth:
            OAuth2 object used by the session to authorize requests.
        :param session:
            The session object to use. If None is provided then an instance of :class:`AuthorizedSession` will be used.
        """
        super().__init__()
        self._oauth = oauth
        if session is not None:
            self._session = session
        else:
            session = session or self.unauthorized_session_class()
            self._session = self.authorized_session_class(self._oauth, **session.get_constructor_kwargs())

    @property
    def auth(self) -> 'OAuth2':
        """
        Get the :class:`OAuth2` instance the client is using for auth to Box.
        """
        return self._oauth

    @property
    def session(self) -> Session:
        """
        Get the :class:`BoxSession` instance the client is using.
        """
        return self._session

    @property
    def translator(self) -> 'Translator':
        """The translator used for translating Box API JSON responses into `BaseAPIJSONObject` smart objects.
        """
        return self._session.translator

    def folder(self, folder_id: str) -> 'Folder':
        """
        Initialize a :class:`Folder` object, whose box id is folder_id.

        :param folder_id:
            The box id of the :class:`Folder` object. Can use '0' to get the root folder on Box.
        :return:
            A :class:`Folder` object with the given folder id.
        """
        return self.translator.get('folder')(session=self._session, object_id=folder_id)

    def root_folder(self) -> 'Folder':
        """
        Returns a user's root folder object.
        """
        return self.folder('0')

    def file(self, file_id: str) -> 'File':
        """
        Initialize a :class:`File` object, whose box id is file_id.

        :param file_id:
            The box id of the :class:`File` object.
        :return:
            A :class:`File` object with the given file id.
        """
        return self.translator.get('file')(session=self._session, object_id=file_id)

    def file_request(self, request_id: str) -> 'FileRequest':
        """
        Initialize a :class:`FileRequest` object, whose box id is request_id.

        :param request_id:
            The box id of the :class:`FileRequest` object.
        :return:
            A :class:`FileRequest` object with the given file request id.
        """
        return self.translator.get('file_request')(session=self._session, object_id=request_id)

    def file_version(self, version_id: str) -> 'FileVersion':
        """
        Initialize a :class:`FileVersion` object, whose box id is version_id.

        :param version_id:
            The box id of the :class:`FileVersion` object.
        :return:
            A :class:`FileVersion` object with the given file version id.
        """
        return self.translator.get('file_version')(session=self._session, object_id=version_id)

    def upload_session(self, session_id: str) -> 'UploadSession':
        """
        Initialize a :class:`UploadSession` object, whose box id is session_id.

        :param session_id:
            The box id of the :class:`UploadSession` object.
        :return:
            A :class:`UploadSession` object with the given session id.
        """
        return self.translator.get('upload_session')(session=self._session, object_id=session_id)

    def comment(self, comment_id: str) -> 'Comment':
        """
        Initialize a :class:`Comment` object, whose Box ID is comment_id.

        :param comment_id:
            The Box ID of the :class:`Comment` object.
        :return:
            A :class:`Comment` object with the given comment ID.
        """
        return self.translator.get('comment')(session=self._session, object_id=comment_id)

    def user(self, user_id: str = 'me') -> 'User':
        """
        Initialize a :class:`User` object, whose box id is user_id.

        :param user_id:
            The user id of the :class:`User` object. Can use 'me' to get the User for the current/authenticated user.
        :return:
            A :class:`User` object with the given id.
        """
        return self.translator.get('user')(session=self._session, object_id=user_id)

    def invite(self, invite_id: str) -> 'Invite':
        """
        Initialize a :class:`Invite` object, whose box id is invite_id.

        :param invite_id:
            The invite ID of the :class:`Invite` object.
        :return:
            A :class:`Invite` object with the given entry ID.
        """
        return self.translator.get('invite')(session=self._session, object_id=invite_id)

    def email_alias(self, alias_id: str) -> 'EmailAlias':
        """
        Initialize a :class: `EmailAlias` object, whose box id is alias_id.

        :param alias_id:
            The aliad id of the :class:`EmailAlias` object.
        :return:
            A :class:`EmailAlias` object with the given entry ID.
        """
        return self.translator.get('email_alias')(session=self._session, object_id=alias_id)

    def group(self, group_id: str) -> 'Group':
        """
        Initialize a :class:`Group` object, whose box id is group_id.

        :param group_id:
            The box id of the :class:`Group` object.
        :return:
            A :class:`Group` object with the given group id.
        """
        return self.translator.get('group')(session=self._session, object_id=group_id)

    def collaboration(self, collab_id: str) -> 'Collaboration':
        """
        Initialize a :class:`Collaboration` object, whose box id is collab_id.

        :param collab_id:
            The box id of the :class:`Collaboration` object.
        :return:
            A :class:`Collaboration` object with the given group id.
        """
        return self.translator.get('collaboration')(session=self._session, object_id=collab_id)

    def collaboration_allowlist(self):
        """
        Initilializes a :class:`CollaborationAllowlist` object.

        :return:
            A :class:`CollaborationAllowlist` object.
        """
        return CollaborationAllowlist(self._session)

    def collaboration_allowlist_entry(self, entry_id: str) -> 'CollaborationAllowlistEntry':
        """
        Initialize a :class:`CollaborationAllowlistEntry` object, whose box id is entry_id.

        :param entry_id:
            The box id of the :class:`CollaborationAllowlistEntry` object.
        :return:
            A :class:`CollaborationAllowlistEntry` object with the given entry id.
        """
        return self.translator.get('collaboration_whitelist_entry')(session=self._session, object_id=entry_id)

    def collaboration_allowlist_exempt_target(self, exemption_id: str) -> 'CollaborationAllowlistExemptTarget':
        """
        Initialize a :class:`CollaborationAllowlistExemptTarget` object, whose box id is target_id.

        :param exemption_id:
            The box id of the :class:`CollaborationAllowlistExemptTarget` object.
        :return:
            A :class:`CollaborationAllowlistExemptTarget` object with the given target id.
        """
        return self.translator.get('collaboration_whitelist_exempt_target')(
            session=self._session,
            object_id=exemption_id
        )

    def trash(self) -> Trash:
        """
        Initialize a :class:`Trash` object.

        :return:
            A :class:`Trash` object.
        """
        return Trash(self._session)

    def legal_hold_policy(self, policy_id: str) -> 'LegalHoldPolicy':
        """
        Initialize a :class:`LegalHoldPolicy` object, whose box id is policy_id.

        :param policy_id:
            The box ID of the :class:`LegalHoldPolicy` object.
        :return:
            A :class:`LegalHoldPolicy` object with the given entry ID.
        """
        return self.translator.get('legal_hold_policy')(session=self._session, object_id=policy_id)

    def legal_hold_policy_assignment(self, policy_assignment_id: str) -> 'LegalHoldPolicyAssignment':
        """
        Initialize a :class:`LegalHoldPolicyAssignment` object, whose box id is policy_assignment_id.

        :param policy_assignment_id:
            The assignment ID of the :class:`LegalHoldPolicyAssignment` object.
        :return:
            A :class:`LegalHoldPolicyAssignment` object with the given entry ID.
        """
        return self.translator.get('legal_hold_policy_assignment')(session=self._session, object_id=policy_assignment_id)

    def legal_hold(self, hold_id: str) -> 'LegalHold':
        """
        Initialize a :class:`LegalHold` object, whose box id is policy_id.

        :param hold_id:
            The legal hold ID of the :class:`LegalHold` object.
        :return:
            A :class:`LegalHold` object with the given entry ID.
        """
        return self.translator.get('legal_hold')(session=self._session, object_id=hold_id)

    @api_call
    def create_legal_hold_policy(
            self,
            policy_name: str,
            description: Optional[str] = None,
            filter_started_at: Union[datetime, str] = None,
            filter_ended_at: Union[datetime, str] = None,
            is_ongoing: Optional[bool] = None
    ) -> 'LegalHoldPolicy':
        """
        Create a legal hold policy.

        :param policy_name:
            The legal hold policy's display name.
        :param description:
            The description of the legal hold policy.
        :param filter_started_at:
            The start date filter for legal hold policy. Takes a datetime string supported by the dateutil library
            or a datetime.datetime object. If no timezone info provided, local timezone will be applied.
        :param filter_ended_at:
            The end date filter for legal hold policy. Takes a datetime string supported by the dateutil library
            or a datetime.datetime object. If no timezone info provided, local timezone will be applied.
        :param is_ongoing:
            After initialization, Assignments under this Policy will continue applying to
            files based on events, indefinitely.
        :returns:
            A legal hold policy object
        """
        url = self.get_url('legal_hold_policies')
        policy_attributes = {'policy_name': policy_name}
        if description is not None:
            policy_attributes['description'] = description
        if filter_started_at is not None:
            policy_attributes['filter_started_at'] = normalize_date_to_rfc3339_format(filter_started_at)
        if filter_ended_at is not None:
            policy_attributes['filter_ended_at'] = normalize_date_to_rfc3339_format(filter_ended_at)
        if is_ongoing is not None:
            policy_attributes['is_ongoing'] = is_ongoing
        box_response = self._session.post(url, data=json.dumps(policy_attributes))
        response = box_response.json()
        return self.translator.translate(
            session=self._session,
            response_object=response,
        )

    @api_call
    def get_legal_hold_policies(
            self,
            policy_name: Optional[str] = None,
            limit: Optional[int] = None,
            marker: Optional[str] = None,
            fields: Iterable[str] = None
    ) -> 'BoxObjectCollection':
        """
        Get the entries in the legal hold policy using limit-offset paging.

        :param policy_name:
            The name of the legal hold policy case insensitive to search for
        :param limit:
            The maximum number of entries to return per page. If not specified, then will use the server-side default.
        :param marker:
            The paging marker to start paging from.
        :param fields:
            List of fields to request.
        :returns:
            An iterator of the entries in the legal hold policy
        """
        additional_params = {}
        if policy_name is not None:
            additional_params['policy_name'] = policy_name
        return MarkerBasedObjectCollection(
            session=self._session,
            url=self.get_url('legal_hold_policies'),
            additional_params=additional_params,
            limit=limit,
            marker=marker,
            fields=fields,
            return_full_pages=False,
        )

    def collection(self, collection_id: str) -> 'Collection':
        """
        Initialize a :class:`Collection` object, whose box ID is collection_id.

        :param collection_id:
            The box id of the :class:`Collection` object.
        :return:
            A :class:`Collection` object with the given collection ID.
        """
        return self.translator.get('collection')(session=self._session, object_id=collection_id)

    @api_call
    def collections(
            self,
            limit: Optional[int] = None,
            offset: int = 0,
            fields: Iterable[str] = None
    ) -> 'BoxObjectCollection':
        """
        Get a list of collections for the current user.

        :param limit:
            The maximum number of users to return. If not specified, the Box API will determine an appropriate limit.
        :param offset:
            The user index at which to start the response.
        :param fields:
            List of fields to request.
        """
        return LimitOffsetBasedObjectCollection(
            self.session,
            self._session.get_url('collections'),
            limit=limit,
            fields=fields,
            offset=offset,
            return_full_pages=False,
        )

    def enterprise(self, enterprise_id: str) -> 'Enterprise':
        """
        Initialize a :class:`Enterprise` object, whose box ID is enterprise_id.

        :param enterprise_id:
            The box id of the :class:`Enterprise` object.
        :return:
            A :class:`Enterprise` object with the given enterprise ID.
        """
        return self.translator.get('enterprise')(session=self._session, object_id=enterprise_id)

    @api_call
    def get_current_enterprise(self) -> 'Enterprise':
        """
        Get the enterprise of the current user.

        :returns:
            The authenticated user's enterprise
        """
        user = self.user().get(fields=['enterprise'])
        enterprise_object = user['enterprise']
        return self.translator.translate(
            session=self._session,
            response_object=enterprise_object,
        )

    @api_call
    def users(
            self,
            limit: Optional[int] = None,
            offset: int = 0,
            filter_term: Optional[str] = None,
            user_type: Optional[str] = None,
            fields: Iterable[str] = None,
            use_marker: bool = False,
            marker: Optional[str] = None
    ) -> Iterable['User']:
        """
        Get a list of all users for the Enterprise along with their user_id, public_name, and login.

        :param limit:
            The maximum number of users to return. If not specified, the Box API will determine an appropriate limit.
        :param offset:
            The user index at which to start the response.
        :param filter_term:
            Filters the results to only users starting with the filter_term in either the name or the login.
        :param user_type:
            Filters the results to only users of the given type: 'managed', 'external', or 'all'.
        :param fields:
            List of fields to request on the :class:`User` objects.
        :param use_marker:
            Whether to use marker-based paging instead of offset-based paging, defaults to False.
        :param marker:
            The paging marker to start returning items from when using marker-based paging.
        :return:
            The list of all users in the enterprise.
        """
        url = self.get_url('users')
        additional_params = {}
        if filter_term:
            additional_params['filter_term'] = filter_term
        if user_type:
            additional_params['user_type'] = user_type

        if use_marker:
            additional_params['usemarker'] = True
            return MarkerBasedObjectCollection(
                url=url,
                session=self._session,
                limit=limit,
                marker=marker,
                fields=fields,
                additional_params=additional_params,
                return_full_pages=False,
            )
        return LimitOffsetBasedObjectCollection(
            url=url,
            session=self._session,
            additional_params=additional_params,
            limit=limit,
            offset=offset,
            fields=fields,
            return_full_pages=False,
        )

    @api_call
    def search(self) -> Search:
        """
        Get a Search object that can be used for searching Box content.

        :return:
            The Search object
        """
        return Search(self._session)

    def events(self) -> Events:
        """
        Get an events object that can get the latest events from Box or set up a long polling event subscription.
        """
        return Events(self._session)

    def group_membership(self, group_membership_id: str) -> 'GroupMembership':
        """
        Initialize a :class:`GroupMembership` object, whose box id is group_membership_id.

        :param group_membership_id:
            The box id of the :class:`GroupMembership` object.
        :return:
            A :class:`GroupMembership` object with the given membership id.
        """
        return self.translator.get('group_membership')(
            session=self._session,
            object_id=group_membership_id,
        )

    @api_call
    def get_groups(
            self,
            name: Optional[str] = None,
            limit: Optional[int] = None,
            offset: Optional[int] = None,
            fields: Iterable[str] = None
    ) -> Iterable['Group']:
        """
        Get a list of all groups for the current user.

        :param name:
            Filter on the name of the groups to return.
        :param limit:
            The maximum number of groups to return. If not specified, the Box API will determine an appropriate limit.
        :param offset:
            The group index at which to start the response.
        :param fields:
            List of fields to request on the :class:`Group` objects.
        :return:
            The collection of all groups.
        """
        url = self.get_url('groups')
        additional_params = {}
        if name:
            additional_params['filter_term'] = name
        return LimitOffsetBasedObjectCollection(
            url=url,
            session=self._session,
            additional_params=additional_params,
            limit=limit,
            offset=offset,
            fields=fields,
            return_full_pages=False,
        )

    def webhook(self, webhook_id: str) -> 'Webhook':
        """
        Initialize a :class:`Webhook` object, whose box id is webhook_id.

        :param webhook_id:
            The box ID of the :class: `Webhook` object.
        :return:
            A :class:`Webhook` object with the given entry ID.
        """
        return self.translator.get('webhook')(session=self._session, object_id=webhook_id)

    @api_call
    def create_webhook(self, target: Union['File', 'Folder'], triggers: Union[list, str], address: str) -> 'Webhook':
        """
        Create a webhook on the given file.

        :param target:
            Either a :class:`File` or :class:`Folder` to assign a webhook to.
        :param triggers:
            Event types that trigger notifications for the target.
        :param address:
            The url to send the notification to.
        :return:
            A :class:`Webhook` object with the given entry ID.
        """
        url = self.get_url('webhooks')
        webhook_attributes = {
            'target': {
                'type': target.object_type,
                'id': target.object_id,
            },
            'triggers': triggers,
            'address': address,
        }
        box_response = self._session.post(url, data=json.dumps(webhook_attributes))
        response = box_response.json()
        return self.translator.translate(
            session=self._session,
            response_object=response,
        )

    @api_call
    def get_webhooks(
            self,
            limit: Optional[int] = None,
            marker: Optional[str] = None,
            fields: Iterable[str] = None
    ) -> 'BoxObjectCollection':
        """
        Get all webhooks in an enterprise.

        :param limit:
            The maximum number of entries to return.
        :param marker:
            The position marker at which to begin the response.
        :param fields:
            List of fields to request on the file or folder which the `RecentItem` references.
        :returns:
            An iterator of the entries in the webhook
        """
        return MarkerBasedObjectCollection(
            session=self._session,
            url=self.get_url('webhooks'),
            limit=limit,
            marker=marker,
            fields=fields,
        )

    @api_call
    def create_group(
            self,
            name: str,
            provenance: Optional[str] = None,
            external_sync_identifier: Optional[str] = None,
            description: Optional[str] = None,
            invitability_level: str = None,
            member_viewability_level: str = None,
            fields: Iterable[str] = None,
    ) -> 'Group':
        """
        Create a group with the given name.

        :param name:
            The name of the group.
        :param provenance:
            Used to track the external source where the group is coming from.
        :param external_sync_identifier:
            Used as a group identifier for groups coming from an external source.
        :param description:
            Description of the group.
        :param invitability_level:
            Specifies who can invite this group to folders.
        :param member_viewability_level:
            Specifies who can view the members of this group.
        :param fields:
            List of fields to request on the :class:`Group` objects.
        :return:
            The newly created Group.
        :raises:
            :class:`BoxAPIException` if current user doesn't have permissions to create a group.
        """
        url = self.get_url('groups')
        additional_params = {}
        body_attributes = {
            'name': name,
        }
        if provenance is not None:
            body_attributes['provenance'] = provenance
        if external_sync_identifier is not None:
            body_attributes['external_sync_identifier'] = external_sync_identifier
        if description is not None:
            body_attributes['description'] = description
        if invitability_level is not None:
            body_attributes['invitability_level'] = invitability_level
        if member_viewability_level is not None:
            body_attributes['member_viewability_level'] = member_viewability_level
        if fields is not None:
            additional_params['fields'] = ','.join(fields)
        box_response = self._session.post(url, data=json.dumps(body_attributes), params=additional_params)
        response = box_response.json()
        return self.translator.translate(
            session=self._session,
            response_object=response,
        )

    def storage_policy(self, policy_id: str) -> 'StoragePolicy':
        """
        Initialize a :class:`StoragePolicy` object, whose box id is policy_id.

        :param policy_id:
            The box ID of the :class:`StoragePolicy` object.
        :return:
            A :class:`StoragePolicy` object with the given entry ID.
        """
        return self.translator.get('storage_policy')(session=self._session, object_id=policy_id)

    def storage_policy_assignment(self, assignment_id: str) -> 'StoragePolicyAssignment':
        """
        Initialize a :class:`StoragePolicyAssignment` object, whose box id is assignment_id.

        :param assignment_id:
            The box ID of the :class:`StoragePolicyAssignment` object.
        :return:
            A :class:`StoragePolicyAssignment` object with the given entry ID.
        """
        return self.translator.get('storage_policy_assignment')(session=self._session, object_id=assignment_id)

    def get_storage_policies(
            self,
            limit: Optional[int] = None,
            marker: Optional[str] = None,
            fields: Iterable[str] = None
    ) -> 'BoxObjectCollection':
        """
        Get the entries in the storage policy using marker-based paging.

        :param limit:
            The maximum number of items to return.
        :param marker:
            The paging marker to start returning items from when using marker-based paging.
        :param fields:
            List of fields to request.
        :returns:
            Returns the storage policies available for the current enterprise.
        """
        return MarkerBasedObjectCollection(
            session=self._session,
            url=self.get_url('storage_policies'),
            limit=limit,
            marker=marker,
            fields=fields,
            return_full_pages=False,
        )

    def terms_of_service(self, tos_id: str) -> 'TermsOfService':
        """
        Initialize a :class:`TermsOfService` object, whose box id is tos_id.

        :param tos_id:
            The box id of the :class:`TermsOfService` object.
        :return:
            A :class:`TermsOfService` object with the given terms of service id.
        """
        return self.translator.get('terms_of_service')(session=self._session, object_id=tos_id)

    def terms_of_service_user_status(self, tos_user_status_id: str) -> 'TermsOfServiceUserStatus':
        """
        Initialize a :class:`TermsOfServiceUserStatus` object, whose box id is tos_user_status_id.

        :param tos_user_status_id:
            The box id of the :class:`TermsOfServiceUserStatus` object.
        :return:
            A :class:`TermsOfServiceUserStatus` object with the given terms of service user status id.
        """
        return self.translator.get('terms_of_service_user_status')(session=self._session, object_id=tos_user_status_id)

    def get_terms_of_services(
            self,
            tos_type: 'TermsOfServiceType' = None,
            limit: Optional[int] = None,
            fields: Iterable[str] = None
    ) -> 'BoxObjectCollection':
        """
        Get the entries in the terms of service using limit-offset paging.
        :param tos_type:
            Can be set to `managed` or `external` for the type of terms of service.
        :param limit:
            The maximum number of items to return. If limit is set to None, then the default
            limit (returned by Box in the response) is used.
        :param fields:
            List of fields to request
        :returns:
            An iterator of the entries in the terms of service
        """
        additional_params = {}
        if tos_type is not None:
            additional_params['tos_type'] = tos_type
        return MarkerBasedObjectCollection(
            session=self._session,
            url=self._session.get_url('terms_of_services'),
            additional_params=additional_params,
            limit=limit,
            marker=None,
            fields=fields,
            return_full_pages=False,
        )

    def task(self, task_id: str) -> 'Task':
        """
        Initialize a :class:`Task` object, whose box id is task_id.

        :param task_id:
            The box ID of the :class:`Task` object.
        :return:
            A :class:`Task` object with the given entry ID.
        """
        return self.translator.get('task')(session=self._session, object_id=task_id)

    def task_assignment(self, assignment_id: str) -> 'TaskAssignment':
        """
        Initialize a :class:`TaskAssignment` object, whose box id is assignment_id.

        :param assignment_id:
            The box ID of the :class:`TaskAssignment` object.
        :return:
            A :class:`TaskAssignment` object with the given entry ID.
        """
        return self.translator.get('task_assignment')(session=self._session, object_id=assignment_id)

    def retention_policy(self, retention_id: str) -> 'RetentionPolicy':
        """
        Initialize a :class:`RetentionPolicy` object, whose box id is retention_id.

        :param retention_id:
            The box ID of the :class:`RetentionPolicy` object.
        :return:
            A :class:`RetentionPolicy` object with the given entry ID.
        """
        return self.translator.get('retention_policy')(session=self._session, object_id=retention_id)

    def file_version_retention(self, retention_id: str) -> 'FileVersionRetention':
        """
        Initialize a :class:`FileVersionRetention` object, whose box id is retention_id.

        :param retention_id:
            The box ID of the :class:`FileVersionRetention` object.
        :return:
            A :class:`FileVersionRetention` object with the given retention ID.
        """
        return self.translator.get('file_version_retention')(session=self._session, object_id=retention_id)

    def retention_policy_assignment(self, assignment_id: str) -> 'RetentionPolicyAssignment':
        """
        Initialize a :class:`RetentionPolicyAssignment` object, whose box id is assignment_id.

        :param assignment_id:
            The box ID of the :class:`RetentionPolicyAssignment` object.
        :return:
            A :class:`RetentionPolicyAssignment` object with the given assignment ID.
        """
        return self.translator.get('retention_policy_assignment')(session=self._session, object_id=assignment_id)

    @api_call
    def create_retention_policy(
            self,
            policy_name: str,
            disposition_action: str,
            retention_length: Union[int, float],
            can_owner_extend_retention: Optional[bool] = None,
            are_owners_notified: Optional[bool] = None,
            custom_notification_recipients: Iterable['User'] = None,
            retention_type: Optional[str] = None,
            description: Optional[str] = None,
    ) -> 'RetentionPolicy':
        """
        Create a retention policy for the given enterprise.

        :param policy_name:
            The name of the retention policy.
        :param disposition_action:
            For `finite` policy can be set to `permanently delete` or `remove retention`.
            For `indefinite` policy this must be set to `remove_retention`
        :param retention_length:
            The amount of time in days to apply the retention policy to the selected content.
            The retention_length should be set to float('inf') for indefinite policies.
        :param can_owner_extend_retention:
            The owner of a file will be allowed to extend the retention if set to true.
        :param are_owners_notified:
            The owner or co-owner will get notified when a file is nearing expiration.
        :param custom_notification_recipients:
            A custom list of user mini objects that should be notified when a file is nearing expiration.
        :param retention_type:
            Specifies the retention type. It can be one of the values:
            -   `modifiable`: You can modify the retention policy. For example, you can add or remove folders,
                shorten or lengthen the policy duration, or delete the assignment.
            -   `non_modifiable`: You can modify the retention policy only in a limited way: add a folder,
                lengthen the duration, retire the policy, change the disposition action or notification settings.
                You cannot perform other actions, such as deleting the assignment or shortening the policy duration.
        :param description:
            The additional text description of the retention policy.
        :return:
            The newly created Retention Policy
        """
        url = self.get_url('retention_policies')
        retention_attributes = {
            'policy_name': policy_name,
            'disposition_action': disposition_action,
        }
        if retention_length == float('inf'):
            retention_attributes['policy_type'] = 'indefinite'
        else:
            retention_attributes['policy_type'] = 'finite'
            retention_attributes['retention_length'] = retention_length
        if can_owner_extend_retention is not None:
            retention_attributes['can_owner_extend_retention'] = can_owner_extend_retention
        if are_owners_notified is not None:
            retention_attributes['are_owners_notified'] = are_owners_notified
        if custom_notification_recipients is not None:
            user_list = [{'type': user.object_type, 'id': user.object_id} for user in custom_notification_recipients]
            retention_attributes['custom_notification_recipients'] = user_list
        if retention_type is not None:
            retention_attributes['retention_type'] = retention_type
        if description is not None:
            retention_attributes['description'] = description
        box_response = self._session.post(url, data=json.dumps(retention_attributes))
        response = box_response.json()
        return self.translator.translate(
            session=self._session,
            response_object=response
        )

    @api_call
    def get_retention_policies(
            self,
            policy_name: Optional[str] = None,
            policy_type: Optional[str] = None,
            user: Optional['User'] = None,
            limit: Optional[int] = None,
            marker: Optional[str] = None,
            fields: Iterable[str] = None,
    ) -> 'BoxObjectCollection':
        """
        Get the entries in the retention policy using marker-based paging.

        :param policy_name:
            The name of the retention policy.
        :param policy_type:
            Set to either `finite` or `indefinite`
        :param user:
            A user to filter the retention policies.
        :param limit:
            The maximum number of entries to return per page. If not specified, then will use the server-side default.
        :param marker:
            The paging marker to start paging from
        :param fields:
            List of fields to request
        :returns:
            An iterator of the entries in the retention policy
        """
        additional_params = {}
        if policy_name is not None:
            additional_params['policy_name'] = policy_name
        if policy_type is not None:
            additional_params['policy_type'] = policy_type
        if user is not None:
            additional_params['created_by_user_id'] = user.object_id
        return MarkerBasedObjectCollection(
            session=self._session,
            url=self._session.get_url('retention_policies'),
            additional_params=additional_params,
            limit=limit,
            marker=marker,
            fields=fields,
            return_full_pages=False,
        )

    def create_terms_of_service(
            self,
            status: 'TermsOfServiceStatus',
            tos_type: 'TermsOfServiceType',
            text: str
    ) -> 'TermsOfService':
        """
        Create a terms of service.

        :param status:
            The status of the terms of service.
        :param tos_type:
            The type of the terms of service. Can be set to `managed` or `external`.
        :param text:
            The message of the terms of service.
        :returns:
            A newly created :class:`TermsOfService` object
        """
        url = self.get_url('terms_of_services')
        body = {
            'status': status,
            'tos_type': tos_type,
            'text': text
        }
        box_response = self._session.post(url, data=json.dumps(body))
        response = box_response.json()
        return self.translator.translate(
            session=self._session,
            response_object=response,
        )

    @deprecated("Use RetentionPolicyAssignment.get_files_under_retention "
                "or RetentionPolicyAssignment.get_file_versions_under_retention instead")
    @api_call
    def get_file_version_retentions(
            self,
            target_file: Optional['File'] = None,
            file_version: Optional['FileVersion'] = None,
            policy: Optional['RetentionPolicy'] = None,
            disposition_action: Optional[str] = None,
            disposition_before: Optional[str] = None,
            disposition_after: Optional[str] = None,
            limit: Optional[int] = None,
            marker: Optional[str] = None,
            fields: Iterable[str] = None,
    ) -> 'BoxObjectCollection':
        """
        Get the entries in the file version retention.

        :param target_file:
            The file to filter the file version.
        :param file_version:
            A file version to filter the file version retentions by.
        :param policy:
            A policy to filter the file version retentions by.
        :param disposition_action:
            Can be set to `permanently_delete` or `remove_retention`.
        :param disposition_before:
            A date time filter for disposition action.
        :param disposition_after:
            A date time filter for disposition action.
        :param limit:
            The maximum number of entries to return per page. If not specified, then will use the server-side default.
        :param marker:
            The paging marker to start paging from
        :param fields:
            List of fields to request
        :returns:
           An iterator of the entries in the file version retention.
        """
        additional_params = {}
        if target_file is not None:
            additional_params['file_id'] = target_file.object_id
        if file_version is not None:
            additional_params['file_version_id'] = file_version.object_id
        if policy is not None:
            additional_params['policy_id'] = policy.object_id
        if disposition_action is not None:
            additional_params['disposition_action'] = disposition_action
        if disposition_before is not None:
            additional_params['disposition_before'] = disposition_before
        if disposition_after is not None:
            additional_params['disposition_after'] = disposition_after
        return MarkerBasedObjectCollection(
            session=self._session,
            url=self._session.get_url('file_version_retentions'),
            additional_params=additional_params,
            limit=limit,
            marker=marker,
            fields=fields,
            return_full_pages=False,
        )

    def web_link(self, web_link_id: str) -> 'WebLink':
        """
        Initialize a :class: `WebLink` object, whose box id is web_link_id.
        :param web_link_id:
            The box ID of the :class:`WebLink` object.
        :return:
            A :class:`WebLink` object with the given entry ID.
        """
        return self.translator.get('web_link')(session=self._session, object_id=web_link_id)

    @api_call
    def get_recent_items(
            self,
            limit: Optional[int] = None,
            marker: Optional[str] = None,
            fields: Iterable[str] = None,
            **collection_kwargs: Any
    ) -> MarkerBasedObjectCollection:
        """
        Get the user's recently accessed items.

        :param limit:
            The maximum number of items to return. If limit is set to None, then the default
            limit (returned by Box in the response) is used. See https://developer.box.com/en/reference/get-recent-items/
            for default.
        :param marker:
            The index at which to start returning items.
        :param fields:
            List of fields to request on the file or folder which the `RecentItem` references.
        :param collection_kwargs:
            Keyword arguments passed to `MarkerBasedObjectCollection`.
        :returns:
            An iterator on the user's recent items
        """
        return MarkerBasedObjectCollection(
            self.session,
            self.get_url('recent_items'),
            limit=limit,
            fields=fields,
            marker=marker,
            **collection_kwargs
        )

    @api_call
    def get_shared_item(self, shared_link: str, password: str = None) -> 'Item':
        """
        Get information about a Box shared link. https://developer.box.com/en/reference/get-shared-items/

        :param shared_link:
            The shared link.
        :param password:
            The password for the shared link.
        :return:
            The item referred to by the shared link.
        :raises:
            :class:`BoxAPIException` if current user doesn't have permissions to view the shared link.
        """
        response = self.make_request(
            'GET',
            self.get_url('shared_items'),
            headers=get_shared_link_header(shared_link, password),
        ).json()
        return self.translator.translate(
            session=self._session.with_shared_link(shared_link, password),
            response_object=response,
        )

    @api_call
    def make_request(self, method: str, url: str, **kwargs: Any) -> 'BoxResponse':
        """
        Make an authenticated request to the Box API.

        :param method:
            The HTTP verb to use for the request.
        :param url:
            The URL for the request.
        :return:
            The network response for the given request.
        :raises:
            :class:`BoxAPIException`
        """
        return self._session.request(method, url, **kwargs)

    @api_call
    def create_user(self, name: str, login: Optional[str] = None, **user_attributes: Any) -> 'User':
        """
        Create a new user. Can only be used if the current user is an enterprise admin, or the current authorization
        scope is a Box developer edition instance.

        :param name:
            The user's display name.
        :param login:
            The user's email address. Required for an enterprise user, but None for an app user.
        :param user_attributes:
            Additional attributes for the user. See the documentation at
            https://developer.box.com/en/reference/post-users/
        :return
            Newly created user
        """
        url = self.get_url('users')
        user_attributes['name'] = name
        if login is not None:
            user_attributes['login'] = login
        else:
            user_attributes['is_platform_access_only'] = True
        box_response = self._session.post(url, data=json.dumps(user_attributes))
        response = box_response.json()
        return self.translator.translate(
            session=self._session,
            response_object=response,
        )

    @api_call
    def get_pending_collaborations(
            self,
            limit: Optional[int] = None,
            offset: Optional[int] = None,
            fields: Iterable[str] = None
    ) -> 'BoxObjectCollection':
        """
        Get the entries in the pending collaborations using limit-offset paging.

        :param limit:
            The maximum number of entries to return per page. If not specified, then will use the server-side default.
        :param offset:
            The offset of the item at which to begin the response.
        :param fields:
            List of fields to request.
        :returns:
            An iterator of the entries in the pending collaborations
        """
        return LimitOffsetBasedObjectCollection(
            session=self._session,
            url=self.get_url('collaborations'),
            additional_params={'status': 'pending'},
            limit=limit,
            offset=offset,
            fields=fields,
            return_full_pages=False,
        )

    @api_call
    def downscope_token(
            self,
            scopes: Iterable['TokenScope'],
            item: 'Item' = None,
            additional_data: dict = None,
            shared_link: str = None
    ) -> TokenResponse:
        """
        Generate a downscoped token for the provided file or folder with the provided scopes.
        There is no way of refreshing this token. You will need to add you own logic to
        generate new token if this one expires.

        :param scopes:
            The scope(s) to apply to the resulting token.
        :param item:
            (Optional) The file or folder to get a downscoped token for. If None and shared_link None, the resulting
            token will not be scoped down to just a single item.
        :param additional_data:
            (Optional) Key value pairs which can be used to add/update the default data values in the request.
        :param shared_link:
            (Optional) The shared link to get a downscoped token for. If None and item None, the resulting token
            will not be scoped down to just a single item.
        :return:
            The response for the downscope token request.
        """
        url = f'{self._session.api_config.OAUTH2_API_URL}/token'
        access_token = self.auth.access_token or self.auth.refresh(None)
        data = {
            'subject_token': access_token,
            'subject_token_type': 'urn:ietf:params:oauth:token-type:access_token',
            'scope': ' '.join(scopes),
            'grant_type': 'urn:ietf:params:oauth:grant-type:token-exchange',
        }

        if item:
            data['resource'] = item.get_url()
        if shared_link:
            data['box_shared_link'] = shared_link
        if additional_data:
            data.update(additional_data)

        box_response = self._session.post(url, data=data)

        return TokenResponse(box_response.json())

    def clone(self, session: 'Session' = None) -> 'Client':
        """Base class override."""
        return self.__class__(oauth=self._oauth, session=(session or self._session))

    def get_url(self, endpoint: str, *args: Any) -> str:
        """
        Return the URL for the given Box API endpoint.

        :param endpoint:
            The name of the endpoint.
        :param args:
            Additional parts of the endpoint URL.
        """
        # pylint:disable=no-self-use
        return self._session.get_url(endpoint, *args)

    def device_pinner(self, device_pin_id: str) -> 'DevicePinner':
        """
        Initialize a :class:`DevicePinner` object, whose box id is device_pin_id.

        :param device_pin_id:
            The assignment ID of the :class:`DevicePin` object.
        :return:
            A :class:`DevicePinner` object with the given entry ID.
        """
        return self.translator.get('device_pinner')(session=self._session, object_id=device_pin_id)

    def device_pinners(
            self,
            enterprise: Optional['Enterprise'] = None,
            direction: Optional[str] = None,
            limit: Optional[int] = None,
            marker: Optional[str] = None,
            fields: Iterable[str] = None
    ) -> 'BoxObjectCollection':
        """
        Returns all of the device pins for the given enterprise.

        :param enterprise:
            The enterprise to retrieve device pinners for, defaulting to the current enterprise.
        :param direction:
            The sorting direction. Set to `ASC` or `DESC`
        :param limit:
            The maximum number of entries to return per page. If not specified, then will use the server-side default.
        :param marker:
            The paging marker to start paging from.
        :param fields:
            List of fields to request.
        :returns:
            An iterator of the entries in the device pins.
        """
        enterprise_id = enterprise.object_id if enterprise is not None else self.get_current_enterprise().id
        additional_params = {}
        if direction is not None:
            additional_params['direction'] = direction
        return MarkerBasedObjectCollection(
            session=self._session,
            url=self.get_url('enterprises', enterprise_id, 'device_pinners'),
            additional_params=additional_params,
            limit=limit,
            marker=marker,
            fields=fields,
            return_full_pages=False,
        )

    def metadata_cascade_policy(self, policy_id: str) -> 'MetadataCascadePolicy':
        """
        Initializes a :class:`MetadataCascadePolicy` object with the given policy ID.

        :param policy_id:
            The ID of the cascade policy object
        :returns:
            The cascade policy object
        """
        return self.translator.get('metadata_cascade_policy')(
            session=self._session,
            object_id=policy_id,
        )

    def metadata_template(self, scope: str, template_key: str) -> 'MetadataTemplate':
        """
        Initialize a :class:`MetadataTemplate` object with the given scope and template key.

        :param scope:
            The scope of the metadata template, e.g. 'enterprise' or 'global'
        :param template_key:
            The key of the metadata template
        :returns:
            The metadata template object
        """
        return self.translator.get('metadata_template')(
            session=self._session,
            object_id=None,
            response_object={
                'type': 'metadata_template',
                'scope': scope,
                'templateKey': template_key,
            },
        )

    def metadata_template_by_id(self, template_id: str) -> 'MetadataTemplate':
        """
        Retrieves a metadata template by ID

        :param template_id:
            The ID of the template object
        :returns:
            The metadata template with data populated from the API
        """
        return self.translator.get('metadata_template')(
            session=self._session,
            object_id=template_id,
        )

    @api_call
    def get_metadata_templates(
            self,
            scope: str = 'enterprise',
            limit: Optional[int] = None,
            marker: Optional[str] = None,
            fields: Iterable[str] = None
    ) -> 'BoxObjectCollection':
        """
        Get all metadata templates for a given scope.  By default, retrieves all metadata templates for the current
        enterprise.

        :param scope:
            The scope to retrieve templates for
        :param limit:
            The maximum number of entries to return per page.
        :param marker:
            The paging marker to start paging from.
        :param fields:
            List of fields to request.
        :returns:
            The collection of metadata templates for the given scope
        """
        return MarkerBasedObjectCollection(
            url=self._session.get_url('metadata_templates', scope),
            session=self._session,
            limit=limit,
            marker=marker,
            fields=fields,
            return_full_pages=False,
        )

    @api_call
    def create_metadata_template(
            self,
            display_name: str,
            fields: Iterable['MetadataField'],
            template_key: str = None,
            hidden: bool = False,
            scope: str = 'enterprise',
            copy_instance_on_item_copy: bool = False
    ) -> 'MetadataTemplate':
        """
        Create a new metadata template.  By default, only the display name and fields are required; the template key
        will be automatically generated based on the display name and the template will be created in the enterprise
        scope.

        :param display_name:
            The human-readable name of the template
        :param fields:
            The metadata fields for the template.
        :param template_key:
            An optional key for the template.  If one is not provided, it will be derived from the display name.
        :param hidden:
            Whether the template should be hidden in the UI
        :param scope:
            The scope the template should be created in
        :param copy_instance_on_item_copy:
            Whether or not to include the metadata when a file or folder is copied.
        """
        url = self._session.get_url('metadata_templates', 'schema')
        body = {
            'scope': scope,
            'displayName': display_name,
            'hidden': hidden,
            'fields': [field.json() for field in fields],
            'copyInstanceOnItemCopy': copy_instance_on_item_copy
        }

        if template_key is not None:
            body['templateKey'] = template_key

        response = self._session.post(url, data=json.dumps(body)).json()
        return self.translator.translate(
            session=self._session,
            response_object=response,
        )

    @api_call
    def __create_zip(self, name: str, items: Iterable) -> dict:
        """
        Creates a zip file containing multiple files and/or folders for later download.

        :param name:
            The name of the zip file to be created.
        :param items:
            List of files and/or folders to be contained in the zip file.
        :returns:
            A dictionary representing a created zip
        """
        # pylint: disable=protected-access
        url = self._session.get_url('zip_downloads')
        zip_file_items = []
        for item in items:
            zip_file_items.append({'type': item._item_type, 'id': item.object_id})
        data = {
            'download_file_name': name,
            'items': zip_file_items
        }
        return self._session.post(url, data=json.dumps(data), skip_retry_codes={202}).json()

    @api_call
    def download_zip(self, name: str, items: Iterable, writeable_stream: IO) -> dict:
        """
        Downloads a zip file containing multiple files and/or folders.

        :param name:
            The name of the zip file to be created.
        :param items:
            List of files or folders to be part of the created zip.
        :param writeable_stream:
            Stream to pipe the readable stream of the zip file.
        :returns:
            A status response object
        """
        created_zip = self.__create_zip(name, items)
        response = self._session.get(created_zip['download_url'], expect_json_response=False, stream=True)
        for chunk in response.network_response.response_as_stream.stream(decode_content=True):
            writeable_stream.write(chunk)
        status = self._session.get(created_zip['status_url']).json()
        status.update(created_zip)
        return self.translator.translate(
            session=self._session,
            response_object=status,
        )

    def folder_lock(self, folder_lock_id: str) -> 'FolderLock':
        """
        Initialize a :class:`FolderLock` object, whose box id is folder_lock_id.

        :param folder_lock_id:
            The ID of the :class:`FolderLock` object.
        :return:
            A :class:`FolderLock` object with the given entry ID.
        """
        return self.translator.get('folder_lock')(session=self._session, object_id=folder_lock_id)

    def sign_request(self, sign_request_id: str) -> 'SignRequest':
        """
        Initialize a :class:`SignRequest` object, whose box id is sign_request_id.

        :param sign_request_id:
            The box id of the :class:`SignRequest` object.
        :return:
            A :class:`SignRequest` object with the given file id.
        """
        return self.translator.get('sign_request')(session=self._session, object_id=sign_request_id)

    @api_call
    def create_sign_request(
            self,
            files: Iterable,
            signers: Iterable,
            parent_folder_id: str,
            prefill_tags: Optional[Iterable] = None,
            are_reminders_enabled: Optional[bool] = None,
            are_text_signatures_enabled: Optional[bool] = None,
            days_valid: Optional[str] = None,
            email_message: Optional[Iterable] = None,
            email_subject: Optional[str] = None,
            external_id: Optional[str] = None,
            is_document_preparation_needed: Optional[bool] = None,
            redirect_url: Optional[str] = None,
            declined_redirect_url: Optional[str] = None,
            template_id: Optional[str] = None,
    ) -> dict:
        """
        Used to create a new sign request.

        :param files:
            List of files to create a signing document from.
        :param signers:
            List of signers for the sign request. 35 is the max number of signers permitted.
        :param parent_folder_id:
            The id of the destination folder to place sign request specific data in.
        :param prefill_tags:
            When a document contains sign related tags in the content,
            you can prefill them using this prefill_tags by referencing the 'id' of the tag as the external_id field of the prefill tag.
        :param are_reminders_enabled:
            Reminds signers to sign a document on day 3, 8, 13 and 18. Reminders are only sent to outstanding signers.
        :param are_text_signatures_enabled:
            Disables the usage of signatures generated by typing (text).
        :param days_valid:
            Number of days after which this request will automatically expire if not completed.
        :param email_message:
            Message to include in sign request email. The field is cleaned through sanitization of specific characters.
            However, some html tags are allowed. Links included in the message are also converted to hyperlinks in the email.
            The message may contain the following html tags including a, abbr, acronym, b, blockquote, code, em, i, ul, li, ol, and strong.
            Be aware that when the text to html ratio is too high, the email may end up in spam filters. Custom styles on these tags are not allowed.
            If this field is not passed, a default message will be used.
        :param email_subject:
            Subject of sign request email. This is cleaned by sign request. If this field is not passed, a default subject will be used.
        :param external_id:
            This can be used to reference an ID in an external system that the sign request is related to.
        :param is_document_preparation_needed:
            Indicates if the sender should receive a prepare_url in the response to complete document preparation via UI.
        :param redirect_url:
            The URL that a signer will be redirected to after signing a document.
            If no declined redirect URL is specified, this URL will be used for decline actions as well.
        :param declined_redirect_url:
            The URL that a signer will be redirected to after declining to sign a document.
        :param template_id:
            The ID of the sign template to use for the sign request.
        :returns:
            A dictionary representing a created SignRequest
        """
        url = self._session.get_url('sign_requests')

        body = {
            'source_files': files,
            'signers': signers,
            'parent_folder': {
                'id': parent_folder_id,
                'type': 'folder'
            }
        }

        if prefill_tags:
            body['prefill_tags'] = prefill_tags
        if are_reminders_enabled:
            body['are_reminders_enabled'] = are_reminders_enabled
        if are_text_signatures_enabled:
            body['are_text_signatures_enabled'] = are_text_signatures_enabled
        if days_valid:
            body['days_valid'] = days_valid
        if email_message:
            body['email_message'] = email_message
        if email_subject:
            body['email_subject'] = email_subject
        if external_id:
            body['external_id'] = external_id
        if is_document_preparation_needed:
            body['is_document_preparation_needed'] = is_document_preparation_needed
        if redirect_url:
            body['redirect_url'] = redirect_url
        if declined_redirect_url:
            body['declined_redirect_url'] = declined_redirect_url
        if template_id:
            body['template_id'] = template_id

        box_response = self._session.post(url, data=json.dumps(body))
        response = box_response.json()
        return self.translator.translate(
            session=self._session,
            response_object=response,
        )

    @api_call
    def get_sign_requests(
            self,
            limit: Optional[int] = None,
            marker: Optional[str] = None,
            fields: Iterable[str] = None
    ) -> 'BoxObjectCollection':
        """
        Returns all the sign requests.

        :param limit:
            The maximum number of entries to return per page. If not specified, then will use the server-side default.
        :param marker:
            The paging marker to start paging from.
        :param fields:
            List of fields to request.
        :returns:
            An iterator of the entries in the device pins.
        """
        return MarkerBasedObjectCollection(
            session=self._session,
            url=self.get_url('sign_requests'),
            limit=limit,
            marker=marker,
            fields=fields,
            return_full_pages=False,
        )

    def sign_template(self, sign_template_id: str) -> 'SignTemplate':
        """
        Initialize a :class:`SignTemplate` object, whose box id is sign_template_id.

        :param sign_template_id:
            The box id of the :class:`SignTemplate` object.
        :return:
            A :class:`SignTemplate` object with the given sign_template_id.
        """
        return self.translator.get('sign_template')(session=self._session, object_id=sign_template_id)

    @api_call
    def get_sign_templates(
            self,
            limit: Optional[int] = None,
            marker: Optional[str] = None,
    ) -> 'BoxObjectCollection':
        """
        Returns all sign templates

        :param limit:
            The maximum number of entries to return per page. If not specified, then will use the server-side default.
        :param marker:
            The paging marker to start paging from.
        :returns:
            Sign templates
        """
        return MarkerBasedObjectCollection(
            session=self._session,
            url=self.get_url("sign_templates"),
            limit=limit,
            marker=marker,
            return_full_pages=False,
        )

    @api_call
    def get_sign_template(
            self,
            sign_template_id: str,
    ) -> Any:
        """
        Returns a sign template

        :param sign_template_id:
            ID of Sign template to fetch
        :returns:
            Sign template
        """
        response = self._session.get(f"{self._session.get_url('sign_templates')}/{sign_template_id}")
        return self.translator.translate(
            session=self._session,
            response_object=response.json(),
        )



from unittest.mock import Mock
import json
import pytest

from boxsdk.config import API
from boxsdk.object.task import Task
from boxsdk.object.task_assignment import TaskAssignment
from boxsdk.network.default_network import DefaultNetworkResponse


@pytest.fixture(scope='module')
def delete_task_response():
    # pylint:disable=redefined-outer-name
    mock_network_response = Mock(DefaultNetworkResponse)
    mock_network_response.ok = True
    return mock_network_response


def test_get(test_task, mock_box_session):
    expected_url = f'{API.BASE_API_URL}/tasks/{test_task.object_id}'
    due_at = '2014-04-03T11:09:43-07:00'
    action = 'review'
    message = 'Test Message'
    mock_box_session.get.return_value.json.return_value = {
        'type': test_task.object_type,
        'id': test_task.object_id,
        'due_at': due_at,
        'action': action,
        'message': message,
    }
    retrieved_task = test_task.get()
    mock_box_session.get.assert_called_once_with(expected_url, headers=None, params=None)
    assert isinstance(retrieved_task, Task)
    assert retrieved_task.object_type == test_task.object_type
    assert retrieved_task.object_id == test_task.object_id
    assert retrieved_task.due_at == due_at
    assert retrieved_task.action == action
    assert retrieved_task.message == message


def test_update(test_task, mock_box_session):
    new_message = 'New Message'
    expected_url = f'{API.BASE_API_URL}/tasks/{test_task.object_id}'
    mock_box_session.put.return_value.json.return_value = {
        'type': test_task.object_type,
        'id': test_task.object_id,
        'message': new_message,
    }
    expected_body = {
        'message': new_message,
    }
    updated_task = test_task.update_info(data={'message': new_message})
    mock_box_session.put.assert_called_once_with(expected_url, data=json.dumps(expected_body), headers=None, params=None)
    assert isinstance(updated_task, Task)
    assert updated_task.message == new_message
    assert updated_task.object_type == test_task.object_type
    assert updated_task.object_id == test_task.object_id


def test_delete_policy_return_the_correct_response(
        test_task,
        mock_box_session,
        delete_task_response,
):
    # pylint:disable=redefined-outer-name
    mock_box_session.delete.return_value = delete_task_response
    response = test_task.delete()
    # pylint:disable=protected-access
    expected_url = f'{API.BASE_API_URL}/tasks/{test_task.object_id}'
    # pylint:enable = protected-access
    mock_box_session.delete.assert_called_once_with(expected_url, params={}, expect_json_response=False, headers=None)
    assert response is True


def test_assign(test_task, mock_user, mock_box_session):
    expected_url = f'{API.BASE_API_URL}/task_assignments'
    expected_body = {
        'task': {
            'type': test_task.object_type,
            'id': test_task.object_id,
        },
        'assign_to': {
            'id': mock_user.object_id,
        },
    }
    mock_box_session.post.return_value.json.return_value = {
        'type': 'task_assignment',
        'id': '42',
        'assigned_to': {
            'type': 'user',
            'id': '1234',
        },
        'assigned_at': '2013-05-10T11:43:41-07:00',
    }
    new_task_assignment = test_task.assign(assignee=mock_user)
    mock_box_session.post.assert_called_once_with(expected_url, data=json.dumps(expected_body))
    assert isinstance(new_task_assignment, TaskAssignment)
    assert new_task_assignment.object_type == 'task_assignment'
    assert new_task_assignment.object_id == '42'
    assert new_task_assignment.assigned_to['type'] == 'user'
    assert new_task_assignment.assigned_at == '2013-05-10T11:43:41-07:00'


def test_assign_with_login(test_task, mock_box_session):
    expected_url = f'{API.BASE_API_URL}/task_assignments'
    expected_body = {
        'task': {
            'type': test_task.object_type,
            'id': test_task.object_id,
        },
        'assign_to': {
            'login': 'test_user@example.com',
        },
    }
    mock_box_session.post.return_value.json.return_value = {
        'type': 'task_assignment',
        'id': '42',
        'assigned_to': {
            'type': 'user',
            'id': '1234',
        },
        'assigned_at': '2013-05-10T11:43:41-07:00',
    }
    new_task_assignment = test_task.assign_with_login(assignee_login='test_user@example.com')
    mock_box_session.post.assert_called_once_with(expected_url, data=json.dumps(expected_body))
    assert isinstance(new_task_assignment, TaskAssignment)
    assert new_task_assignment.object_type == 'task_assignment'
    assert new_task_assignment.object_id == '42'
    assert new_task_assignment.assigned_to['type'] == 'user'
    assert new_task_assignment.assigned_at == '2013-05-10T11:43:41-07:00'


def test_get_assignments(test_task, mock_box_session):
    expected_url = f'{API.BASE_API_URL}/tasks/{test_task.object_id}/assignments'
    mock_assignment = {
        'type': 'task_assignment',
        'id': '12345',
        'item': {
            'type': 'file',
            'id': '33333'
        }
    }
    mock_box_session.get.return_value.json.return_value = {
        'limit': 100,
        'entries': [mock_assignment]
    }
    assignments = test_task.get_assignments()
    assignment = assignments.next()
    mock_box_session.get.assert_called_once_with(expected_url, params={})
    assert isinstance(assignment, TaskAssignment)
    assert assignment.id == mock_assignment['id']
    assert assignment.item['id'] == mock_assignment['item']['id']



import warnings
from typing import Callable, Any


def deprecated(message: str):
    def deprecated_decorator(func: Callable) -> Any:
        def deprecated_func(*args, **kwargs):
            warnings.simplefilter('default', DeprecationWarning)
            warnings.warn(f'{func.__name__} is a deprecated function. {message}',
                          category=DeprecationWarning,
                          stacklevel=2)
            return func(*args, **kwargs)
        return deprecated_func
    return deprecated_decorator


def deprecated_param(*, name: str, position: int, message: str):
    def deprecated_decorator(func: Callable) -> Any:
        def deprecated_func(*args, **kwargs):
            if len(args) >= position + 1 or name in kwargs:
                warnings.simplefilter('default', DeprecationWarning)
                warnings.warn(
                    f'{func.__name__} function parameter `{name}` at position {position} is deprecated. {message}',
                    category=DeprecationWarning,
                    stacklevel=2
                )
            return func(*args, **kwargs)
        return deprecated_func
    return deprecated_decorator



from unittest.mock import Mock
import json
import pytest

from boxsdk.config import API
from boxsdk.object.storage_policy_assignment import StoragePolicyAssignment
from boxsdk.network.default_network import DefaultNetworkResponse


@pytest.fixture(scope='module')
def delete_assignment_response():
    # pylint:disable=redefined-outer-name
    mock_network_response = Mock(DefaultNetworkResponse)
    mock_network_response.ok = True
    return mock_network_response


def test_get(test_storage_policy_assignment, mock_box_session):
    expected_url = f'{API.BASE_API_URL}/storage_policy_assignments/{test_storage_policy_assignment.object_id}'
    mock_box_session.get.return_value.json.return_value = {
        'type': 'storage_policy_assignment',
        'id': test_storage_policy_assignment.object_id,
    }
    storage_policy_assignment = test_storage_policy_assignment.get()
    mock_box_session.get.assert_called_once_with(expected_url, headers=None, params=None)
    assert isinstance(storage_policy_assignment, StoragePolicyAssignment)


def test_update(test_storage_policy_assignment, mock_box_session):
    new_policy_id = '1234'
    expected_url = f'{API.BASE_API_URL}/storage_policy_assignments/{test_storage_policy_assignment.object_id}'
    mock_box_session.put.return_value.json.return_value = {
        'type': 'storage_policy_assignment',
        'id': new_policy_id,
    }
    storage_policy_assignment = test_storage_policy_assignment.update_info(data={
        'storage_policy': {
            'type': 'storage_policy',
            'id': new_policy_id,
        }
    })
    data = {
        'storage_policy': {
            'type': 'storage_policy',
            'id': new_policy_id,
        }
    }
    mock_box_session.put.assert_called_once_with(expected_url, data=json.dumps(data), headers=None, params=None)
    assert isinstance(storage_policy_assignment, StoragePolicyAssignment)


def test_delete(test_storage_policy_assignment, delete_assignment_response, mock_box_session):
    mock_box_session.delete.return_value = delete_assignment_response
    response = test_storage_policy_assignment.delete()
    expected_url = test_storage_policy_assignment.get_url()
    mock_box_session.delete.assert_called_once_with(expected_url, params={}, expect_json_response=False, headers=None)
    assert response is True



from typing import Optional, Iterator, TYPE_CHECKING

from .box_object_collection import BoxObjectCollection

if TYPE_CHECKING:
    from boxsdk.session.session import Session


class MarkerBasedObjectCollection(BoxObjectCollection):
    """
    An iterator of Box objects (BaseObjects) that were retrieved from a Box API endpoint that supports
    marker type of pagination.

    See https://developer.box.com/en/guides/api-calls/pagination/ for more details.
    """

    def __init__(
            self,
            session: 'Session',
            url: str,
            limit: Optional[int] = None,
            fields: Iterator[str] = None,
            additional_params: Optional[dict] = None,
            return_full_pages: bool = False,
            marker: Optional[str] = None,
            supports_limit_offset_paging: bool = False,
            use_post: bool = False
    ):
        """
        :param session:
            The Box session used to make requests.
        :param url:
            The endpoint url to hit.
        :param limit:
            The number of entries for each page to return. The default, as well as the upper limit of this value,
            differs by endpoint. See https://developer.box.com/en/reference. If limit is set to None, then the default
            limit (returned by Box in the response) is used.
        :param fields:
            List of fields to request. If None, will return the default fields for the object.
        :param additional_params:
            Additional HTTP params to send in the request.
        :param return_full_pages:
            If True, then the returned iterator for this collection will return full pages of Box objects on each
            call to next(). If False, the iterator will return a single Box object on each next() call.
        :param marker:
            The offset index to start paging from.
        :param supports_limit_offset_paging:
            Does this particular endpoint also support limit-offset paging? This information is needed, as
            the endpoints that support both require an special extra request parameter.
        :param use_post:
            If True, then the returned iterator will make POST requests with all the data in the body on each
            call to next().
            If False, the iterator will make GET requets with all the data as query params on each call to next().
        """
        super().__init__(
            session,
            url,
            limit=limit,
            fields=fields,
            additional_params=additional_params,
            return_full_pages=return_full_pages,
            use_post=use_post
        )
        self._marker = marker
        self._supports_limit_offset_paging = supports_limit_offset_paging

    def _update_pointer_to_next_page(self, response_object: dict) -> None:
        """Baseclass override."""
        self._marker = self._get_next_marker_from_response_object(response_object)

    def _has_more_pages(self, response_object) -> bool:
        """Baseclass override."""
        return bool(self._get_next_marker_from_response_object(response_object))

    @staticmethod
    def _get_next_marker_from_response_object(response_object) -> Optional[str]:
        """Get the marker that should be used to retrieve the next page.

        When we've just retrieved the last page, the API is inconsistent about
        what it returns. Some endpoints return "next_marker":"", some return
        "next_marker":null, some don't give any "next_marker" value. In all of
        these cases, this method will return `None`.

        Otherwise, this method returns the string value of the "next_marker"
        field.
        """
        return response_object.get('next_marker') or None

    def _next_page_pointer_params(self) -> dict:
        """Baseclass override."""
        pointer_params = {}
        # For transitioning endpoints that support both marker and limit-offset paging, we must specify an
        # additional 'useMarker' parameter to the Box API.
        if self._supports_limit_offset_paging:
            pointer_params['useMarker'] = True
        if self._marker is not None:
            pointer_params['marker'] = self._marker
        return pointer_params

    def next_pointer(self) -> Optional[str]:
        """Baseclass override."""
        return self._marker



from typing import Optional

import attr

from .network.network_interface import NetworkResponse
from .util.log import sanitize_dictionary


class BoxException(Exception):
    """
    Base class exception for all errors raised from the SDK.
    """
    def __str__(self):
        return self.__class__.__name__

    def __repr__(self):
        return f'<{self.__class__.__name__}>'


class BoxValueError(ValueError):
    """
    Exception raise by SDK value errors
    """


class BoxNetworkException(BoxException):
    """
    Exception raised from the network layer.
    """


@attr.s(repr=True, slots=True, frozen=True)
class BoxAPIException(BoxException):
    """
    Exception raised from the box session layer.

    :param status:
        HTTP status code of the failed response
    :param code:
        The 'code' field of the failed response
    :param message:
        A message to associate with the exception, e.g. 'message' field of the json in the failed response
    :param request_id:
        The 'request_id' field of the json in the failed response
    :param headers:
        The HTTP headers in the failed response
    :param url:
        The url which raised the exception
    :param method:
        The HTTP verb used to make the request.
    :param context_info:
        The context_info returned in the failed response.
    :param network_response:
        The failed response
    """
    status: int = attr.ib()
    code: Optional[str] = attr.ib(default=None)
    message: Optional[str] = attr.ib(default=None)
    request_id: Optional[str] = attr.ib(default=None)
    headers: dict = attr.ib(default=None, hash=False)
    url: str = attr.ib(default=None)
    method: str = attr.ib(default=None)
    context_info: Optional[dict] = attr.ib(default=None)
    network_response: 'NetworkResponse' = attr.ib(default=None, repr=False)

    def __str__(self):
        return '\n'.join((
            f'Message: {self.message}',
            f'Status: {self.status}',
            f'Code: {self.code}',
            f'Request ID: {self.request_id}',
            f'Headers: {sanitize_dictionary(self.headers)}',
            f'URL: {self.url}',
            f'Method: {self.method}',
            f'Context Info: {self.context_info}',
        ))


@attr.s(repr=True, slots=True, frozen=True)
class BoxOAuthException(BoxException):
    """
    Exception raised during auth.

    :param status:
        HTTP status code of the auth response
    :param message:
        A message to associate with the exception, e.g. HTTP content of the auth response
    :param url:
        The url which raised the exception
    :param method:
        The HTTP verb used to make the request.
    :param network_response:
        The network response for the request.
    :param code:
        The 'code' field of the failed response
    """
    status: int = attr.ib()
    message: str = attr.ib(default=None)
    url: str = attr.ib(default=None)
    method: str = attr.ib(default=None)
    network_response: NetworkResponse = attr.ib(default=None, repr=False)
    code: Optional[str] = attr.ib(default=None)

    def __str__(self):
        # pylint:disable=no-member
        if self.network_response:
            headers = sanitize_dictionary(self.network_response.headers)
        # pylint:enable=no-member
        else:
            headers = 'N/A'
        return f'\nMessage: {self.message}\nStatus: {self.status}\nURL: {self.url}\nMethod: {self.method}' \
               f'\nHeaders: {headers}'


__all__ = ['BoxException', 'BoxAPIException', 'BoxOAuthException', 'BoxNetworkException']



from bottle import request
from sqlalchemy.orm.exc import NoResultFound
from test.functional.mock_box.db_model.file_model import FileModel
from test.functional.mock_box.db_model.folder_model import FolderModel
from test.functional.mock_box.db_model.token_model import TokenModel
from test.functional.mock_box.util.http_utils import abort


def get_file_by_id(db_session, file_id):
    try:
        return db_session.query(FileModel).filter_by(file_id=file_id).one()
    except NoResultFound:
        abort(404)


def get_folder_by_id(db_session, folder_id):
    try:
        return db_session.query(FolderModel).filter_by(folder_id=folder_id).one()
    except NoResultFound:
        abort(404)


def get_token_record_by_token(db_session, token):
    try:
        return db_session.query(TokenModel).filter_by(token=token).one()
    except NoResultFound:
        abort(401)


def get_user_from_header(db_session):
    token = request.headers.get('Authorization')[7:]
    token_record = get_token_record_by_token(db_session, token)
    user_id = token_record.owned_by
    return user_id



from typing import Any

from boxsdk.object.group import Group
from test.integration_new import util
from test.integration_new import CLIENT


class BoxTestGroup:

    def __init__(self, *, name: str = None):
        if name is None:
            name = util.random_name()

        self._group: Group = CLIENT.create_group(name=name)

    def __enter__(self) -> Group:
        return self._group

    def __exit__(self, exc_type: Any, exc_val: Any, exc_tb: Any) -> None:
        self._group.delete()



from unittest.mock import Mock
import json
import pytest

from boxsdk.config import API
from boxsdk.network.default_network import DefaultNetworkResponse
from boxsdk.object.webhook import Webhook


@pytest.fixture(scope='module')
def delete_webhook_response():
    # pylint:disable=redefined-outer-name
    mock_network_response = Mock(DefaultNetworkResponse)
    mock_network_response.ok = True
    return mock_network_response


def test_delete_webhook_return_the_correct_response(
        test_webhook,
        mock_box_session,
        delete_webhook_response,
):
    # pylint:disable=redefined-outer-name
    mock_box_session.delete.return_value = delete_webhook_response
    response = test_webhook.delete()
    expected_url = f'{API.BASE_API_URL}/webhooks/{test_webhook.object_id}'
    mock_box_session.delete.assert_called_once_with(expected_url, params={}, expect_json_response=False, headers=None)
    assert response is True


def test_get(test_webhook, mock_box_session):
    expected_url = f'{API.BASE_API_URL}/webhooks/{test_webhook.object_id}'
    mock_box_session.get.return_value.json.return_value = {
        'type': test_webhook.object_type,
        'id': test_webhook.object_id,
        'created_at': '2016-05-04T18:51:45-07:00',
        'address': 'https://example.com',
        'triggers': ['FILE.UPLOADED'],
    }
    webhook = test_webhook.get()
    mock_box_session.get.assert_called_once_with(expected_url, headers=None, params=None)
    assert isinstance(webhook, Webhook)
    assert webhook.object_type == test_webhook.object_type
    assert webhook.object_id == test_webhook.object_id
    assert webhook.created_at == '2016-05-04T18:51:45-07:00'
    assert webhook.address == 'https://example.com'
    assert webhook.triggers == ['FILE.UPLOADED']


def test_update(test_webhook, mock_box_session):
    expected_url = f'{API.BASE_API_URL}/webhooks/{test_webhook.object_id}'
    mock_box_session.put.return_value.json.return_value = {
        'type': test_webhook.object_type,
        'id': test_webhook.object_id,
        'address': 'https://testnotification.com',
        'triggers': ['FILE.DOWNLOADED']
    }
    data = {
        'address': 'https://testnotification.com',
        'triggers': ['FILE.DOWNLOADED'],
    }
    webhook = test_webhook.update_info(data=data)
    mock_box_session.put.assert_called_once_with(expected_url, data=json.dumps(data), headers=None, params=None)
    assert isinstance(webhook, Webhook)
    assert webhook.type == test_webhook.object_type
    assert webhook.id == test_webhook.object_id
    assert webhook.address == 'https://testnotification.com'
    assert webhook.triggers == ['FILE.DOWNLOADED']


@pytest.mark.parametrize(
    'signature_version,signature_algorithm,primary_key,secondary_key,expected_result',
    [
        ('1', 'HmacSHA256', 'SamplePrimaryKey', 'SampleSecondaryKey', True),
        ('1', 'HmacSHA256', 'SamplePrimaryKey', None, True),
        ('1', 'HmacSHA256', 'WrongPrimaryKey', 'SampleSecondaryKey', True),
        ('1', 'HmacSHA256', 'WrongPrimaryKey', 'WrongSecondaryKey', False),
        ('1', 'HmacSHA256', None, None, False),
        ('2', 'HmacSHA256', 'SamplePrimaryKey', 'SampleSecondaryKey', False),
        ('1', 'WrongAlgorithm', 'SamplePrimaryKey', 'SampleSecondaryKey', False),
    ]
)
def test_validate_message(signature_version, signature_algorithm, primary_key, secondary_key, expected_result):
    # pylint: disable=C0301
    body = b'{"type":"webhook_event","webhook":{"id":"1234567890"},"trigger":"FILE.UPLOADED","source":{"id":"1234567890","type":"file","name":"Test.txt"}}'
    headers = {
        'box-delivery-id': 'f96bb54b-ee16-4fc5-aa65-8c2d9e5b546f',
        'box-delivery-timestamp': '2020-01-01T00:00:00-07:00',
        'box-signature-algorithm': signature_algorithm,
        'box-signature-primary': '6TfeAW3A1PASkgboxxA5yqHNKOwFyMWuEXny/FPD5hI=',
        'box-signature-secondary': 'v+1CD1Jdo3muIcbpv5lxxgPglOqMfsNHPV899xWYydo=',
        'box-signature-version': signature_version,
    }
    is_validated = Webhook.validate_message(body, headers, primary_key, secondary_key)
    assert is_validated is expected_result



from collections import OrderedDict
from datetime import datetime
from itertools import chain
import json
from typing import Optional, Union
from unittest.mock import Mock
from urllib.parse import urlunsplit, urlencode

import pytest
import pytz
from requests.exceptions import Timeout

from boxsdk.network.default_network import DefaultNetworkResponse
from boxsdk.object.events import Events, EventsStreamType, UserEventsStreamType
from boxsdk.object.event import Event
from boxsdk.object.folder import Folder
from boxsdk.session.box_response import BoxResponse


@pytest.fixture()
def test_events(mock_box_session):
    return Events(mock_box_session)


@pytest.fixture()
def final_stream_position():
    return 1348790499820


@pytest.fixture()
def initial_stream_position():
    return 1348790499819


# pylint:disable=no-member
# pylint isn't currently smart enough to recognize the class member that was
# added by the metaclass, when the metaclass was added by @add_metaclass() /
# with_metaclass().
STREAM_TYPES_AS_ENUM_INSTANCES = list(EventsStreamType.__members__.values())
# pylint:enable=no-member
STREAM_TYPES_AS_STRINGS = list(map(str, STREAM_TYPES_AS_ENUM_INSTANCES))


def test_events_stream_type_extended_enum_class_has_expected_members():
    assert len(STREAM_TYPES_AS_ENUM_INSTANCES) >= 4
    assert len(STREAM_TYPES_AS_STRINGS) >= 4
    assert 'all' in STREAM_TYPES_AS_STRINGS
    assert 'changes' in STREAM_TYPES_AS_STRINGS
    assert 'sync' in STREAM_TYPES_AS_STRINGS
    assert 'admin_logs' in STREAM_TYPES_AS_STRINGS


@pytest.fixture(
    scope='session',
    params=list(chain(
        [None],   # Default behavior of not passing any stream_type
        STREAM_TYPES_AS_ENUM_INSTANCES,   # Passing an enum instance
        STREAM_TYPES_AS_STRINGS,  # Passing an enum value

        # For forwards compatibility, make sure that it works to pass a string
        # value that is not a member of the enum.
        ['future_stream_type'],
    )),
)
def stream_type_param(request) -> Optional[Union[str, EventsStreamType]]:
    """The value to pass as an Event method's stream_type parameter.

    :return:
        The parameter value, or `None` if no value should be passed.
    """
    return request.param


@pytest.fixture()
def expected_stream_type(stream_type_param) -> str:
    """The stream type we expect to use.
    """
    if stream_type_param is None:
        return UserEventsStreamType.ALL
    return stream_type_param


@pytest.fixture()
def stream_type_kwargs(stream_type_param) -> dict:
    """The kwargs for stream_type to pass when invoking a method on `Events`.
    """
    if stream_type_param:
        return {'stream_type': stream_type_param}
    return {}


@pytest.fixture()
def expected_stream_type_params(expected_stream_type) -> OrderedDict:
    """The stream_type-related params that we expect to pass to request methods.
    """
    return OrderedDict(stream_type=expected_stream_type)


@pytest.fixture()
def empty_events_response(final_stream_position):
    # pylint:disable=redefined-outer-name
    mock_box_response = Mock(BoxResponse)
    mock_network_response = Mock(DefaultNetworkResponse)
    mock_box_response.network_response = mock_network_response
    mock_box_response.json.return_value = mock_json = {'next_stream_position': final_stream_position, 'entries': []}
    mock_box_response.content = json.dumps(mock_json).encode()
    mock_box_response.status_code = 200
    mock_box_response.ok = True
    return mock_box_response


@pytest.fixture()
def long_poll_url(test_url, expected_stream_type_params):
    return urlunsplit(('', '', test_url, urlencode(expected_stream_type_params), ''))


@pytest.fixture()
def retry_timeout():
    return 610


@pytest.fixture()
def options_response_entry(long_poll_url, retry_timeout):
    return {'url': long_poll_url, 'retry_timeout': retry_timeout}


@pytest.fixture()
def options_response(options_response_entry, make_mock_box_request):
    # pylint:disable=redefined-outer-name
    mock_box_response, _ = make_mock_box_request(
        response={'entries': [options_response_entry]},
    )
    return mock_box_response


@pytest.fixture()
def new_change_long_poll_response(make_mock_box_request):
    mock_box_response, _ = make_mock_box_request(
        response={'message': 'new_change'},
    )
    return mock_box_response


@pytest.fixture()
def reconnect_long_poll_response(make_mock_box_request):
    mock_box_response, _ = make_mock_box_request(
        response={'message': 'reconnect'},
    )
    return mock_box_response


@pytest.fixture()
def max_retries_long_poll_response(make_mock_box_request):
    mock_box_response, _ = make_mock_box_request(
        response={'message': 'max_retries'},
    )
    return mock_box_response


@pytest.fixture()
def mock_event_json():
    return {
        "type": "event",
        "event_id": "f82c3ba03e41f7e8a7608363cc6c0390183c3f83",
        "source": {
            "type": "folder",
            "id": "11446498",
        },
    }


@pytest.fixture()
def events_response(initial_stream_position, mock_event_json, make_mock_box_request):
    # pylint:disable=redefined-outer-name
    mock_box_response, _ = make_mock_box_request(
        response={"next_stream_position": initial_stream_position, "entries": [mock_event_json]},
    )
    return mock_box_response


def test_get_events(
        test_events,
        mock_box_session,
        events_response,
        stream_type_kwargs,
        expected_stream_type_params,
):
    # pylint:disable=redefined-outer-name
    expected_url = test_events.get_url()
    mock_box_session.get.return_value = events_response
    events = test_events.get_events(**stream_type_kwargs)
    assert 'next_stream_position' in events
    mock_box_session.get.assert_any_call(
        expected_url,
        params={'limit': 100, 'stream_position': 0, **expected_stream_type_params},
    )
    for event, json in zip(events['entries'], events_response.json.return_value['entries']):
        assert isinstance(event, Event)
        assert event.event_id == json['event_id']


@pytest.mark.parametrize("limit", [100, None])
@pytest.mark.parametrize('created_after', [
    '2019-07-01T22:02:24+14:00',
    datetime(2019, 7, 1, 22, 2, 24, tzinfo=pytz.timezone('US/Alaska'))
])
@pytest.mark.parametrize('created_before', [
    '2019-08-07T22:02:24+14:00',
    datetime(2019, 8, 7, 22, 2, 24, tzinfo=pytz.timezone('US/Alaska'))
])
def test_get_admin_events(
        test_events,
        mock_box_session,
        events_response,
        limit,
        created_after,
        created_before
):
    # pylint:disable=redefined-outer-name
    expected_url = test_events.get_url()
    mock_box_session.get.return_value = events_response
    events = test_events.get_admin_events(
        limit=limit,
        stream_position=0,
        created_after=created_after,
        created_before=created_before,
        event_types=['ITEM_CREATE', "LOGIN"],
    )
    expected_params = {
        'stream_position': 0,
        'created_after': '2019-07-01T22:02:24+14:00',
        'created_before': '2019-08-07T22:02:24+14:00',
        'event_type': 'ITEM_CREATE,LOGIN',
        'stream_type': 'admin_logs'
    }
    if limit:
        expected_params['limit'] = limit

    mock_box_session.get.assert_called_with(
        expected_url,
        params=expected_params
    )
    for event, json in zip(events['entries'], events_response.json.return_value['entries']):
        assert isinstance(event, Event)
        assert event.event_id == json['event_id']


@pytest.mark.parametrize("limit", [100, None])
def test_get_admin_events_streaming(
        test_events,
        mock_box_session,
        events_response,
        limit,
):
    # pylint:disable=redefined-outer-name
    expected_url = test_events.get_url()
    mock_box_session.get.return_value = events_response
    events = test_events.get_admin_events_streaming(
        limit=limit,
        stream_position=100,
        event_types=['ITEM_CREATE', "LOGIN"],
    )
    expected_params = {
        'stream_type': 'admin_logs_streaming',
        'stream_position': 100,
        'event_type': 'ITEM_CREATE,LOGIN'
    }
    if limit:
        expected_params['limit'] = limit

    mock_box_session.get.assert_called_with(
        expected_url,
        params=expected_params
    )
    for event, json in zip(events['entries'], events_response.json.return_value['entries']):
        assert isinstance(event, Event)
        assert event.event_id == json['event_id']


def test_get_long_poll_options(
        mock_box_session,
        test_events,
        stream_type_kwargs,
        expected_stream_type_params,
        options_response,
        options_response_entry,
):
    expected_url = test_events.get_url()
    mock_box_session.options.return_value = options_response
    long_poll_options = test_events.get_long_poll_options(**stream_type_kwargs)
    mock_box_session.options.assert_called_with(expected_url, params=expected_stream_type_params)
    assert long_poll_options == options_response_entry


def test_generate_events_with_long_polling(
        test_events,
        mock_box_session,
        events_response,
        empty_events_response,
        initial_stream_position,
        long_poll_url,
        retry_timeout,
        options_response,
        new_change_long_poll_response,
        reconnect_long_poll_response,
        max_retries_long_poll_response,
        mock_event_json,
        stream_type_kwargs,
        expected_stream_type,
        expected_stream_type_params,
):
    # pylint:disable=redefined-outer-name
    expected_url = test_events.get_url()
    mock_box_session.options.return_value = options_response
    mock_box_session.get.side_effect = [
        events_response,  # initial call to get now stream position
        Timeout,
        reconnect_long_poll_response,
        max_retries_long_poll_response,
        new_change_long_poll_response,
        events_response,
        new_change_long_poll_response,
        empty_events_response,
        EscapeGenerator("A fake exception for the session to throw so that the generator won't block forever"),
    ]
    events = test_events.generate_events_with_long_polling(**stream_type_kwargs)

    event = next(events)
    assert isinstance(event, Event)
    assert event.event_id == mock_event_json['event_id']
    assert isinstance(event.source, Folder)
    assert event.source.id == mock_event_json['source']['id']
    with pytest.raises(EscapeGenerator):
        next(events)
    events.close()
    mock_box_session.options.assert_called_with(expected_url, params=expected_stream_type_params)
    mock_box_session.get.assert_any_call(expected_url, params={'stream_position': 'now', 'limit': 0, 'stream_type': expected_stream_type})
    assert '/events' in expected_url
    mock_box_session.get.assert_any_call(
        expected_url,
        params={'limit': 100, 'stream_position': initial_stream_position, **expected_stream_type_params},
    )
    mock_box_session.get.assert_any_call(
        long_poll_url,
        timeout=retry_timeout,
        params={'stream_position': initial_stream_position},
    )


class EscapeGenerator(RuntimeError):
    pass



from test.util.streamable_mock_open import streamable_mock_open
from unittest.mock import patch

import pytest

from boxsdk.client import Client
from boxsdk.exception import BoxAPIException


def test_upload_then_delete(box_client, test_file_path, test_file_content, file_name):
    with patch('boxsdk.object.folder.open', streamable_mock_open(read_data=test_file_content), create=True):
        file_object = box_client.folder('0').upload(test_file_path, file_name)
    assert file_object.delete()
    folder_items = box_client.folder('0').get_items(1)
    item_count = 0
    for _ in folder_items:
        item_count += 1
    assert item_count == 0


def test_create_folder_then_update_info(box_client, folder_name):
    folder = box_client.folder('0').create_subfolder(folder_name)
    assert folder.delete()
    folder_items = box_client.folder('0').get_items(1)
    item_count = 0
    for _ in folder_items:
        item_count += 1
    assert item_count == 0


@pytest.mark.parametrize('constructor', [Client.file, Client.folder])
def test_get_item_info_for_missing_file(box_client, constructor):
    with pytest.raises(BoxAPIException) as exc_info:
        constructor(box_client, '1').delete()
    assert exc_info.value.status == 404
